"""
Method advisors are effectively function advisors and are covered there.
"""
from collections import defaultdict

from ..advisors import filt_block_advisor
from .. import conf
from ..utils import get_nice_str_list, layout_comment as layout

CLASS_XPATH = ('descendant-or-self::ClassDef')

@filt_block_advisor(xpath=CLASS_XPATH, warning=True)
def getters_setters(block_dets, *, repeated_message=False):
    """
    Look for getters and setters and suggest @property if appropriate.
    """
    class_els = block_dets.element.xpath(CLASS_XPATH)
    if not class_els:
        return None
    class_getter_setter_methods = defaultdict(list)
    for class_el in class_els:
        class_name = class_el.get('name')
        method_els = class_el.xpath('body/FunctionDef')
        for method_el in method_els:
            method_name = method_el.get('name')
            if method_name.startswith(('set_', 'get_')):
                class_getter_setter_methods[class_name].append(method_name)
    if not class_getter_setter_methods:
        return None
    brief_msg = layout("""\

        ### Alternative to getters and setters

        """)
    for class_name, method_names in sorted(class_getter_setter_methods.items()):
        multiple = len(method_names) > 1
        if multiple:
            nice_list = get_nice_str_list(method_names, quoter='`')
            brief_msg += layout(f"""\

                Class `{class_name}` has the following methods that look like
                either getters or setters: {nice_list}.
                """)
        else:
            method_type = (
                'getter' if method_name.startswith('get_') else 'setter')
            brief_msg += layout(f"""\

                Class `{class_name}` has a `{method_names[0]}` method that
                looks like a {method_type}.
                """)
    if not repeated_message:
        brief_msg += layout("""\
    
            Python doesn't need getters and setters. Instead, you can use
            properties. These are easily added using decorators e.g.
            `@property`.

            """)
    main_msg = brief_msg
    if not repeated_message:
        tm = '\N{TRADE MARK SIGN}'
        main_msg += (
            layout(f"""\
                A good discussion of getters, setters, and properties can be
                found at <https://www.python-course.eu/python3_properties.php>.

                Getters and setters are usually added in other languages such as
                Java because direct attribute access doesn't give the ability to
                calculate results or otherwise run a process when a value is
                accessed / written.

                And it is common for lots of getters and setters to be added,
                whether or not they are actually needed - Just In Case{tm}. The
                fear is that if you point other code to an attribute, and you
                later need to process the attribute or derive it before it is
                served up or stored, then you'll need to make a breaking change
                to your code. All the client code referencing the attribute will
                have to be rewritten to replace direct access with a reference
                to the appropriate getter or setter. Understandably then the
                inclination is to point to a getter or setter in the first case
                even if it doesn't actually do anything different (for now at
                least) from direct access. Using these getters and setters is
                wasteful, and bloats code unnecessarily, but it avoids the worse
                evil of regularly broken interfaces. The benefit is that you can
                change implementation later if you need to and nothing will
                break. But in Python there is a much better way :-).

                Let's compare the getter / setter approach and the property
                approach.

                #### Using getters / setters

            """)
            +
            layout("""

                class Person:

                    def __init__(self, name):
                        self.__name = name

                    def get_name(self):
                        return self.__name

                    def set_name(self, name):
                        if name is not None:
                            self.__name = name

                """, is_code=True)
            +
            layout("""\

                #### Using properties

                """)
            +
            layout("""

                class Person:

                    def __init__(self, name):
                        self.name = name

                    @property  ## the getter
                    def name(self):
                        return self.__name

                    @name.setter  ## the setter
                    def name(self, name):
                        if name is not None:
                            self.__name = name

                """, is_code=True)
            +
            layout("""\

                We must define the getter earlier in the script than the setter
                because the setter references the getter name. If it is not
                already defined above it we will get a `NameError` because
                Python doesn't yet know the variable_name part of the
                `@<variable_name>.setter`.

                In the `\_\_init\_\_` method we can either directly set the
                "private" variable `self.__name` (note - unenforced in Python)
                or set the public variable `self.name` and pass through the
                setter. Passing through the setter makes more sense - presumably
                there are some smarts in the setter we want applied otherwise we
                wouldn't have gone to the trouble of making it ;-).

                You will have also noticed that the method name is defined twice
                without the second one overwriting the first (the standard
                behaviour). The decorators take care of all that. The only thing
                to remember is to use exactly the same attribute name in three
                places (assuming both getting and setting):

                """)
            +
            layout("""\

                @property
                def here():  ## <== 1
                    ...

                @here.setter  ## <== 2
                def here():  ## <== 3
                    ...

                """, is_code=True)
        )
        if repeated_message:
            extra_msg = ''
        else:
            extra_msg = (
                layout("""\
                    Python also has a `deleter` decorator which handle deletion
                    of the attribute e.g.

                    """)
                +
                layout("""\
                    @name.deleter
                    def name(self):
                        ...
                    """, is_code=True)
            )
        message = {
            conf.BRIEF: brief_msg,
            conf.MAIN: main_msg,
            conf.EXTRA: extra_msg,
        }
        return message

@filt_block_advisor(xpath=CLASS_XPATH, warning=True)
def selfless_methods(block_dets, *, repeated_message=False):
    """
    Look for class methods that don't use self as candidates for @staticmethod
    decorator. Note - not worried about detecting sophisticated cases with
    packing etc although it doesn't have to be named "self".
    """
    class_els = block_dets.element.xpath(CLASS_XPATH)
    if not class_els:
        return None
    class_selfless_methods = defaultdict(list)
    for class_el in class_els:
        class_name = class_el.get('name')
        method_els = class_el.xpath('body/FunctionDef')
        for method_el in method_els:
            method_name = method_el.get('name')
            arg_els = method_el.xpath('args/arguments/args/arg')
            if not arg_els:
                continue
            first_arg_name = arg_els[0].get('arg')
            used_first_names = [
                name_el for name_el in method_el.xpath('descendant::Name')
                if name_el.get('id') == first_arg_name]
            if not used_first_names:
                class_selfless_methods[class_name].append(method_name)
    if not class_selfless_methods:
        return None
    brief_msg = layout("""\

        ### Method doesn't use instance

        """)
    for class_name, method_names in sorted(class_selfless_methods.items()):
        multiple = len(method_names) > 1
        if multiple:
            nice_list = get_nice_str_list(method_names, quoter='`')
            brief_msg += layout(f"""\

                Class `{class_name}` has the following methods that don't use
                the instance (usually called `self`): {nice_list}.
                """)
        else:
            brief_msg += layout(f"""\

                Class `{class_name}` has a `{method_names[0]}` method that
                doesn't use the instance (usually called `self`).
                """)
    if not repeated_message:
        brief_msg += layout("""\

            If a method doesn't use the instance it can be either pulled into a
            function outside the class definition or decorated with
            `@staticmethod`. The decorator stops the method expecting the
            instance object to be supplied as the first argument.

            """)
    main_msg = brief_msg

    if not repeated_message:
        main_msg += (
            layout("""
                For example, instead of:

                """)
            +
            layout("""
                def calc_approx_days(self, years):  ## <== self not used!
                    return round(years * 365.25)

                """, is_code=True)
            +
             layout("""

                you could write:

                """)
            +
            layout("""
                @staticmethod
                def calc_approx_days(years):
                    return round(years * 365.25)
                """, is_code=True)
        )

    if repeated_message:
        extra_msg = ''
    else:
        extra_msg = layout("""\
            It is not obligatory to call the first parameter of a bound method
            `self` but you should call it that unless you have a good reason to
            break convention.
            """)
    message = {
        conf.BRIEF: brief_msg,
        conf.MAIN: main_msg,
        conf.EXTRA: extra_msg,
    }
    return message

@filt_block_advisor(xpath=CLASS_XPATH, warning=True)
def one_method_classes(block_dets, *, repeated_message=False):
    """
    Look for classes with only one method (other than __init__) and suggest a
    simple function as an alternative..
    """
    class_els = block_dets.element.xpath(CLASS_XPATH)
    if not class_els:
        return None
    classes_sole_methods = []
    for class_el in class_els:
        method_els = class_el.xpath('body/FunctionDef')
        method_names = [method_el.get('name') for method_el in method_els]
        non_init_method_names = [method_name for method_name in method_names
            if method_name != '__init__']
        n_non_init_methods = len(non_init_method_names)
        if n_non_init_methods < 2:
            class_name = class_el.get('name')
            try:
                sole_method_name = non_init_method_names.pop()
            except IndexError:
                sole_method_name = None
            classes_sole_methods.append((class_name, sole_method_name))
    if not classes_sole_methods:
        return None
    multi_sole = len(classes_sole_methods) > 1
    class_plural = 'es' if multi_sole else ''
    class_have_has = 'have' if multi_sole else 'has'
    func_plural = 's' if multi_sole else ''
    brief_msg = layout(f"""\

        ### Possible option of converting class{class_plural} to single function{func_plural}

        The following class{class_plural} only {class_have_has} one main
        function at most (excluding `__init__`):

        """)
    for class_name, method_name in classes_sole_methods:
        method2use = (
            f"`{method_name}`" if method_name else 'nothing but `__init__`')
        brief_msg += layout(f"""\
        - {class_name}: {method2use}
        """)
    if repeated_message:
        extra_msg = ''
    else:
        brief_msg += (
            layout(f"""\

                It may be simpler to replace the class{class_plural} with simple
                functions e.g. we could replace:

            """)
            +
            layout('''\
                class GetSquare:

                    def __init__(self, num):
                        self.num = num

                    def get_square(self):
                        """
                        Get square of number
                        """
                        return self.num ** 2
                ''', is_code=True)
            +
            layout(f"""\

                with:

            """)
            +
            layout('''\
                def get_square(num):
                    """
                    Get square of number ...
                    """
                    return self.num ** 2
                ''', is_code=True)
            +
            layout(f"""\

                Sometimes, less is more. Bugs need somewhere to hide and
                unnecessarily verbose code gives them plenty of room to hide.

            """)
        )
        extra_msg = layout("""\
            Python allows procedural, object-oriented, and functional styles of
            programming. Event-based programming is also used in GUI contexts,
            for example. Programmers coming to Python from languages that only
            support object-orientation sometimes overdo the classes when there
            is a simpler, more elegant way of writing readable code in Python.

            If only a simple function is required, then write a simple function.

            Note - there may be exceptions. It has been suggested that the class
            structure can make it easier to test intermediate state rather than
            just function outputs. So, as with most things, it depends.
            """)
    message = {
        conf.BRIEF: brief_msg,
        conf.EXTRA: extra_msg,
    }
    return message



import builtins
import keyword
from textwrap import dedent

from ..conf import STD_LIBS
from ..utils import layout_comment as layout

def is_reserved_name(name):
    is_reserved = name in set(keyword.kwlist + dir(builtins) + STD_LIBS)
    return is_reserved

UNPACKING_COMMENT = (
    dedent(f"""\
        Unpacking is much more pythonic than using indexes to pull a sequence
        apart into names (variables). For example:

        """)
    +
    layout(f"""\
        #### Un-pythonic :-(

        location = (-37, 174, 'Auckland', 'Mt Albert')
        lat = location[0]
        lon = location[1]
        city = location[2]
        suburb = location[3]

        #### Pythonic :-)
        lat, lon, city, suburb = location
        """, is_code=True)
    +
    layout(f"""\

        If you don't need all the values you can indicate which you want to
        ignore or even mop up multiple unused values into a single value using
        an asterisk.

        For example:

        """)
    +
    layout(f"""\
        lat, lon, _city, _suburb = location
        """, is_code=True)
    +
    layout(f"""\

        or:

        """)
    +
    layout(f"""\
        lat, lon, *_ = location
        """, is_code=True)
    +
    layout(f"""\

        or:

        """)
    +
    layout(f"""\
        lat, lon, *unused = location
        """, is_code=True)
    +
    layout(f"""\

        Note - unused, in this case, would be ['Auckland', 'Mt Albert']

        """)
)

GENERAL_COMPREHENSION_COMMENT = layout(f"""\
    Comprehensions are one the great things about Python. To see why, have a
    look at Raymond Hettinger's classic talk "Transforming Code into Beautiful,
    Idiomatic Python" <https://youtu.be/OSGv2VnC0go?t=2738> where he explains
    the rationale. In short, if the goal of your code can be expressed as a
    single English sentence then it might belong on one line. The code should
    say what it is doing more than how it is doing it. Comprehensions are
    declarative and that is A Very Good Thing™.

    Pro tip: don't make comprehensions *in*comprehensions ;-). If your
    comprehension is hard to read it is probably better rewritten as a looping
    structure.
    """)

LIST_COMPREHENSION_COMMENT = (
    layout("""\
        #### Example List Comprehension:
        """)
    +
    layout(f"""\
        names_lengths = [
            len(name)
            for name in ['Tinky Winky', 'Dipsy', 'La La', 'Po']
        ]
        """, is_code=True)
    +
    layout("""\

        produces an ordinary list:

        """)
    +
    layout(str(
        {
            len(name)
            for name in ['Tinky Winky', 'Dipsy', 'La La', 'Po']
        }
        ))
    +
    layout("""\

        It is also possible to add a simple filter using `if`

        """)
    +
    layout(f"""\
        names_lengths = [
            len(name)
            for name in ['Tinky Winky', 'Dipsy', 'La La', 'Po']
            if not name.startswith('T')
        ]
        """, is_code=True)
    +
    layout("""\

        produces an ordinary list:

        """)
    +
    layout(str(
        {
            len(name)
            for name in ['Tinky Winky', 'Dipsy', 'La La', 'Po']
            if not name.startswith('T')
        }
        ))
)

DICT_COMPREHENSION_COMMENT = (
    layout("""\
    #### Example Dictionary Comprehension:
    """)
    +
    layout(f"""\
        country2capital = {{
            country: capital
            for country, capital in [('NZ', 'Wellington'), ('Italy', 'Rome')]
        }}
        """, is_code=True)
    +
    layout("""\

        produces an ordinary dictionary:

        """)
    +
    layout(str(
        {
            country: capital
            for country, capital
            in [('NZ', 'Wellington'), ('Italy', 'Rome')]
        }
        ))
    +
    layout("""\

        It is also possible to add a simple filter using `if`

        """)
    +
    layout(f"""\
        country2capital = {{
            country: capital
            for country, capital in [('NZ', 'Wellington'), ('Italy', 'Rome')]
            if country == 'NZ'
        }}
        """, is_code=True)
    +
    layout("""\

        produces an ordinary dictionary:

        """)
    +
    layout(str(
        {
            country: capital
            for country, capital
            in [('NZ', 'Wellington'), ('Italy', 'Rome')]
            if country == 'NZ'
        }
        ))
)

SET_COMPREHENSION_COMMENT = (
    layout("""\
        #### Example Set Comprehension
        """)
    +
    layout(f"""\
        pets = {{
            pet for _person, pet
            in [('Rachel', 'cat'), ('Elliot', 'goat'), ('Giles', 'cat'),]
        }}
        """, is_code=True)
    +
    layout("""\

        produces an ordinary set (i.e. unique members only):

        """)
    +
    layout(str(
        {
            pet for _person, pet
                in [('Rachel', 'cat'), ('Elliot', 'goat'), ('Giles', 'cat'),]
        }
        ))
    +
    layout("""\

        It is also possible to add a simple filter using `if`

        """)
    +
    layout(f"""\
        pets = {{
            pet for person, pet
            in [('Rachel', 'cat'), ('Elliot', 'goat'), ('Giles', 'cat'),]
            if person != 'Elliot'
        }}
        """, is_code=True)
    +
    layout("""\

        produces an ordinary set (i.e. unique members only):

        """)
    +
    layout(str(
        {
            pet for person, pet
                in [('Rachel', 'cat'), ('Elliot', 'goat'), ('Giles', 'cat'),]
                if person != 'Elliot'
        }
        ))
)



from collections import namedtuple

from ..advisors import snippet_advisor
from .. import conf
from ..utils import layout_comment as layout

NTDets = namedtuple('NamedTupleDetails', 'name, label, fields_str, fields_list')

def get_named_tuple_dets(named_tuple_el):
    """
    Get name, label, fields.
    """
    ancestor_elements = named_tuple_el.xpath('ancestor-or-self::*')
    assign_block_el = ancestor_elements[-5]  ##  Assign value Call func Name
    name = assign_block_el.xpath('targets/Name')[0].get('id')
    label_el, fields_el = assign_block_el.xpath('value/Call/args/Str')
    label = label_el.get('s')
    fields_str = fields_el.get('s')
    fields_list = [field.strip() for field in fields_str.split(',')]
    named_tuple_dets = NTDets(name, label, fields_str, fields_list)
    return named_tuple_dets

def get_named_tuples_dets(blocks_dets):
    all_named_tuples_dets = []
    for block_dets in blocks_dets:
        func_name_els = block_dets.element.xpath(
            'descendant-or-self::value/Call/func/Name')
        named_tuple_els = [func_name_el for func_name_el in func_name_els
            if func_name_el.get('id') == 'namedtuple']
        named_tuples_dets = [
            get_named_tuple_dets(named_tuple_el)
            for named_tuple_el in named_tuple_els]
        all_named_tuples_dets.extend(named_tuples_dets)
    return all_named_tuples_dets

@snippet_advisor()
def named_tuple_overview(blocks_dets):
    """
    Look for named tuples and explain how they can be enhanced.
    """
    named_tuples_dets = get_named_tuples_dets(blocks_dets)
    if not named_tuples_dets:
        return None
    example_dets = named_tuples_dets[0]
    first_field = example_dets.fields_list[0]
    brief_msg = (
            layout("""\

                ### Named Tuple Enhancements

                Named tuples can be enhanced to make them even more useful -
                especially when debugging. The label can be expanded beyond the
                variable name; and the entire named tuple or individual fields
                can be given their own doc strings.

                For example:

                """)
            +
            layout(f"""\
                {example_dets.name} = namedtuple("{example_dets.label}",
                    "{example_dets.fields_str}")
                {example_dets.name}.__doc__ += "\\n\\nExtra comments"
                {example_dets.name}.{first_field}.__doc__ = "Specific comment for {first_field}"
                ## etc
                """, is_code=True)
        )
    main_msg = (
        brief_msg
        +
        layout("""\

            Default arguments are another nice option (added in Python 3.7). For
            example the following named tuple has a default IQ of 100:
            """)
        +
        layout("""\

            People = namedtuple('PeopleDets', 'name, IQ', defaults=(100, ))
            """, is_code=True)
        +
        layout("""\

            The official documentation has more details.
            """)
    )
    message = {
        conf.BRIEF: brief_msg,
        conf.MAIN: main_msg,
    }
    return message
    


"""
Covers functions and methods.
"""
from ..advisors import filt_block_advisor
from ..ast_funcs import get_el_lines_dets
from .. import conf, utils
from ..utils import get_python_version, layout_comment as layout

def get_danger_status_3_7(child_el):
    if (child_el.tag == 'NameConstant'
            and child_el.get('value') in ['True', 'False']):
        danger_status = 'Boolean'
    elif child_el.tag == 'Num' and child_el.get('n'):
        danger_status = 'Number'
    else:
        danger_status = None
    return danger_status

def get_danger_status_3_8(child_el):
    if child_el.tag == 'Constant':
        val = child_el.get('value')
        if val in ['True', 'False']:
            danger_status = 'Boolean'
        else:
            try:
                float(val)
            except TypeError:
                danger_status = None
            else:
                danger_status = 'Number'
    else:
        danger_status = None
    return danger_status

def get_docstring_from_value_3_7(first_value_el):
    if first_value_el.tag != 'Str':
        docstring = None
    else:
        docstring = first_value_el.get('s')
    return docstring

def get_docstring_from_value_3_8(first_value_el):
    if first_value_el.tag != 'Constant':
        docstring = None
    else:
        docstring = first_value_el.get('value')
    return docstring

python_version = get_python_version()
if python_version in (conf.PY3_6, conf.PY3_7):
    get_danger_status = get_danger_status_3_7
    get_docstring_from_value = get_docstring_from_value_3_7
elif python_version == conf.PY3_8:
    get_danger_status = get_danger_status_3_8
    get_docstring_from_value = get_docstring_from_value_3_8
else:
    raise Exception(f"Unexpected Python version {python_version}")

FUNC_DEFN_XPATH = 'descendant-or-self::FunctionDef'

def get_is_method(func_el):
    class_els = func_el.xpath('ancestor::ClassDef')
    if not class_els:
        return False
    class_el = class_els[-1]
    ## check is a direct member of a class and not just a function defined somewhere internally
    method_els = class_el.xpath('body/FunctionDef')
    return func_el in method_els

def get_func_type_lbl(func_el):
    func_type_lbl = (
        conf.METHOD_LBL if get_is_method(func_el) else conf.FUNCTION_LBL)
    return func_type_lbl

def get_overall_func_type_lbl(func_els):
    if not func_els:
        return None
    includes_plain_function = False
    for func_el in func_els:
        func_type_lbl = get_func_type_lbl(func_el)
        if func_type_lbl == conf.FUNCTION_LBL:
            includes_plain_function = True
    overall_func_type_lbl = (
        conf.FUNCTION_LBL if includes_plain_function else conf.METHOD_LBL)
    return overall_func_type_lbl

def _get_arg_comment(func_el, *, repeated_message=False):
    """
    Must cope with positional arguments, keyword arguments, keyword-only
    arguments, packed positional arguments, and packed keyword arguments.
    Trivial really ;-).

    Comment should end without a full stop because calling code adds that to
    make the sentence structure more explicit.
    """
    args = func_el.xpath('args/arguments/args/arg')
    vararg = func_el.xpath('args/arguments/vararg/arg')
    kwarg = func_el.xpath('args/arguments/kwarg/arg')
    kwonlyargs = func_el.xpath('args/arguments/kwonlyargs/arg')
    has_packing = (vararg or kwarg)
    if has_packing:
        arg_comment = 'receives a variable number of arguments'
        if repeated_message:
            arg_comment += '.'
        else:
            if vararg:
                vararg_name = vararg[0].get('arg')
                arg_comment += (". All positional arguments received are packed"
                    f" together into a list called {vararg_name} using "
                    f"the &ast;{vararg_name} syntax. If there is no better name"
                    " in a particular case the Python convention is to call "
                    "that list 'args'")
            if kwarg:
                kwarg_name = kwarg[0].get('arg')
                arg_comment += (". All keyword arguments received are packed "
                    f"together into a dictionary called {kwarg_name} "
                    f"using the &ast;&ast;{kwarg_name} syntax. If there is no "
                    "better name in a particular case the Python convention is "
                    "to call that dictionary 'kwargs'")
    else:
        all_args_n = len(args + kwonlyargs)
        if all_args_n:
            nice_n_args = utils.int2nice(all_args_n)
            arg_comment = (f"receives {nice_n_args} argument")
            if all_args_n > 1:
                arg_comment += 's'
        else:
            arg_comment = "doesn't take any arguments"
    return arg_comment

def _get_return_comment(func_type_lbl, return_elements, *,
        repeated_message=False):
    implicit_return_els = [return_element for return_element in return_elements
        if not return_element.getchildren()]
    implicit_returns_n = len(implicit_return_els)
    explicit_return_els = [return_element for return_element in return_elements
        if return_element.xpath('value')]
    none_returns_n = 0
    val_returns_n = 0
    for el in explicit_return_els:
        val_children = el.getchildren()
        for val_child in val_children:
            if (val_child.tag == 'NameConstant'
                    and val_child.get('value') is None):
                none_returns_n += 1
            else:
                val_returns_n += 1
    keyword_returns_n = none_returns_n + val_returns_n + implicit_returns_n
    if not keyword_returns_n:
        returns_comment = (
            f"The {func_type_lbl} does not explicitly return anything.")
        if not repeated_message:
            returns_comment += (
                " In which case, in Python, it implicitly returns None")
    else:
        returns_comment = (
            f"The {func_type_lbl} exits via an explicit `return` statement "
            f"{utils.int2nice(keyword_returns_n)} time")
        if repeated_message:
            returns_comment += '.'
        else:
            if keyword_returns_n > 1:
                returns_comment += (
                    f"s. Some people prefer {func_type_lbl}s to have one, and "
                    "only one exit point for clarity. Others disagree e.g. "
                    f"using early returns to short-circuit {func_type_lbl}s if "
                    "initial validation of some sort makes it possible to avoid"
                    f" the bulk of the {func_type_lbl}. Whatever approach you "
                    f"take make sure your {func_type_lbl} is easy to reason "
                    "about in terms of what it returns and where it exits")
            else:
                returns_comment += '.'
    return returns_comment

def _get_exit_comment(func_el, func_type_lbl, *, repeated_message=False):
    """
    Look for 'return' and 'yield'.
    """
    return_elements = func_el.xpath('descendant-or-self::Return')
    yield_elements = func_el.xpath('descendant-or-self::Yield')
    if yield_elements:
        if return_elements:
            exit_comment = ("It has both `return` and `yield`. "
                "That probably doesn't make any sense.")
        else:
            exit_comment = "It is a generator function."
    else:
        exit_comment = _get_return_comment(func_type_lbl,
            return_elements, repeated_message=repeated_message)
    return exit_comment

@filt_block_advisor(xpath=FUNC_DEFN_XPATH)
def func_overview(block_dets, *, repeated_message=False):
    """
    Advise on function (or method) definition statements.
    e.g. def greeting(): ...
    """
    func_els = block_dets.element.xpath(FUNC_DEFN_XPATH)
    overall_func_type_lbl = get_overall_func_type_lbl(func_els)
    brief_msg = layout(f"""\
        ### {overall_func_type_lbl.title()} Details
        """)
    for func_el in func_els:
        func_type_lbl = get_func_type_lbl(func_el)
        name = func_el.get('name')
        arg_comment = _get_arg_comment(
            func_el, repeated_message=repeated_message)
        exit_comment = _get_exit_comment(
            func_el, func_type_lbl, repeated_message=repeated_message)
        brief_msg += layout(f"""\

            The {func_type_lbl} named `{name}` {arg_comment}. {exit_comment}.
            """)

    main_msg = brief_msg
    if repeated_message:
        extra_msg = ''
    else:
        if func_type_lbl == conf.METHOD_LBL:
            main_msg += layout("""\

                Methods are functions that sit directly inside a class
                definition. Unless they are defined as static methods e.g. using
                the `@staticmethod` decorator, they take the instance of the
                class as the first parameter - almost always named `self`. But
                they are basically functions.
                """)
        extra_msg = layout(f"""\
            There is often confusion about the difference between arguments and
            parameters. {overall_func_type_lbl.title()}s define parameters but
            receive arguments. You can think of parameters as being like car
            parks and arguments as the cars that fill them. You supply arguments
            to a {overall_func_type_lbl} depending on its parameters.
            """)
    message = {
        conf.BRIEF: brief_msg,
        conf.EXTRA: extra_msg,
    }
    return message

@filt_block_advisor(xpath=FUNC_DEFN_XPATH, warning=True)
def func_len_check(block_dets, *, repeated_message=False):
    """
    Warn about functions that might be too long.
    """
    func_els = block_dets.element.xpath(FUNC_DEFN_XPATH)
    overall_func_type_lbl = get_overall_func_type_lbl(func_els)
    brief_msg = ''
    has_short_comment = False
    for func_el in func_els:
        func_type_lbl = get_func_type_lbl(func_el)
        name = func_el.get('name')
        first_line_no, last_line_no, _func_lines_n = get_el_lines_dets(
            func_el, ignore_trailing_lines=True)
        block_lines = block_dets.block_code_str.split('\n')
        func_lines = block_lines[first_line_no - 1: last_line_no]
        func_non_empty_lines = [line for line in func_lines if line]
        func_lines_n = len(func_non_empty_lines)
        if func_lines_n <= conf.MAX_BRIEF_FUNC_LOC:
            continue
        else:
            if not has_short_comment:
                brief_msg += layout(f"""\
                    ### {overall_func_type_lbl.title()} possibly too long

                    """)
                has_short_comment = True
            brief_msg += layout(f"""\

            `{name}` has {utils.int2nice(func_lines_n)} lines of code
            (including comments but with empty lines ignored).
            """)
            if not repeated_message:
                brief_msg += (
                    f" Sometimes it is OK for a {func_type_lbl} to be "
                    "that long but you should consider refactoring the code "
                    "into smaller units.")
    if not brief_msg:
        return None
    message = {
        conf.BRIEF: brief_msg,
    }
    return message

def get_n_args(func_el):
    arg_els = func_el.xpath('args/arguments/args/arg')
    posonlyarg_els = func_el.xpath('args/arguments/posonlyargs/arg')
    kwonlyarg_els = func_el.xpath('args/arguments/kwonlyargs/arg')
    n_args = len(arg_els + posonlyarg_els + kwonlyarg_els)
    return n_args

@filt_block_advisor(xpath=FUNC_DEFN_XPATH, warning=True)
def func_excess_parameters(block_dets, *, repeated_message=False):
    """
    Warn about functions that might have too many parameters.
    """
    func_els = block_dets.element.xpath(FUNC_DEFN_XPATH)
    brief_msg = ''
    has_high = False
    overall_func_type_lbl = get_overall_func_type_lbl(func_els)
    for func_el in func_els:
        func_type_lbl = get_func_type_lbl(func_el)
        name = func_el.get('name')
        n_args = get_n_args(func_el)
        high_args = n_args > conf.MAX_BRIEF_FUNC_ARGS
        if high_args:
            brief_msg += layout(f"""\
                ### Possibly too many {overall_func_type_lbl} parameters

                """)
            brief_msg += layout(f"""\

            `{name}` has {n_args:,} parameters.

            """)
            if not (has_high or repeated_message):
                brief_msg += layout(f"""\
                    Sometimes it is OK for a {func_type_lbl} to have that many
                    but you should consider refactoring the code or collecting
                    related parameters into single parameters e.g. instead of
                    receiving image size arguments separately perhaps you could
                    receive a dictionary of image size argument details.
                    """)
            has_high = True
    if not has_high:
        return None
    message = {
        conf.BRIEF: brief_msg,
    }
    return message

def get_danger_args(func_el):
    arg_els = func_el.xpath('args/arguments/args/arg')  ## not kwonlyargs so potentially supplied positionally only
    arg_names = [arg_el.get('arg') for arg_el in arg_els]
    arg_default_els = func_el.xpath('args/arguments/defaults')
    danger_statuses = []
    for arg_default_el in arg_default_els:
        for child_el in arg_default_el.getchildren():
            danger_status = get_danger_status(child_el)
            danger_statuses.append(danger_status)
    ## reversed because defaults are filled in rightwards e.g. a, b=1, c=2
    ## args = a,b,c and defaults=1,2 -> reversed c,b,a and 2,1 -> c: 2, b: 1
    arg_names_reversed = reversed(arg_names)
    danger_statuses_reversed = reversed(danger_statuses)
    args_and_danger_statuses = zip(arg_names_reversed, danger_statuses_reversed)
    danger_args = [(arg, danger_status)
        for arg, danger_status in args_and_danger_statuses
        if danger_status is not None]
    return danger_args

@filt_block_advisor(xpath=FUNC_DEFN_XPATH, warning=True)
def positional_boolean(block_dets, *, repeated_message=False):
    """
    Look for any obvious candidates for forced keyword use e.g. where a
    parameter is a boolean or a number.

    Defaults apply from the rightmost backwards (within their group - either
    defaults or kw_defaults (related to kwonlyargs)).
    """
    func_els = block_dets.element.xpath(FUNC_DEFN_XPATH)
    overall_func_type_lbl = get_overall_func_type_lbl(func_els)
    brief_msg = ''
    has_positional_comment = False
    for func_el in func_els:
        func_type_lbl = get_func_type_lbl(func_el)
        name = func_el.get('name')
        danger_args = get_danger_args(func_el)
        if danger_args:
            if not has_positional_comment:
                brief_msg += layout(f"""\

                    ### {overall_func_type_lbl.title()} expects risky positional
                    arguments
                    """)
                if not repeated_message:
                    brief_msg += layout(f"""\

                    {func_type_lbl.title}s which expect numbers or booleans
                    (True/False) without requiring keywords are risky. They are
                    risky when if the {func_type_lbl} is changed later to have
                    different parameters. For example, greeting(formal=True) is
                    more intelligible than greeting(True). And intelligible code
                    is safer to alter / maintain over time than mysterious code.
                    """)
            brief_msg += layout(f"""\
                A partial analysis of `{name}` found the following risky non-
                keyword (positional) parameters: {danger_args}.
                """)
            if not (has_positional_comment) or repeated_message:
                brief_msg += (
                    layout("""\

                        Using an asterisk as a pseudo-parameter forces all
                        parameters to the right to be keywords e.g.
                        """)
                    +
                    layout(f"""\
                        def greeting(name, *, formal=False):
                            ...
                        """, is_code=True)
                    +
                    layout(f"""\

                        In this example you couldn't now call the function
                        greeting('Jo', True) - it would need to be
                        greeting('Jo', formal=True)
                        """)
                )
            has_positional_comment = True
    if not has_positional_comment:
        return None
    if repeated_message:
        extra_msg = ''
    else:
        extra_msg = layout(f"""\
            Putting an asterisk in the parameters has the effect of forcing all
            parameters to the right to be keyword parameters because the
            asterisk mops up any remaining positional arguments supplied (if
            any) when the {func_type_lbl} is called. There can't be any other
            positional arguments, because they have all been handled already, so
            only keyword parameters are allowed thereafter.
            """)
    message = {
        conf.BRIEF: brief_msg,
        conf.EXTRA: extra_msg,
    }
    return message

def get_func_name_docstring(func_el):
    func_body_el = func_el.xpath('body')[0]
    func_name = func_el.get('name')
    ## first item in body MUST be Expr and the first value must be a Str
    body_els = func_body_el.getchildren()
    if not body_els:
        docstring = None
    else:
        first_body_el = body_els[0]
        if first_body_el.tag != 'Expr':
            docstring = None
        else:
            expr_els = first_body_el.getchildren()
            if not expr_els:
                docstring = None
            else:
                first_expr_el = expr_els[0]
                if first_expr_el.tag != 'value':
                    docstring = None
                else:
                    value_els = first_expr_el.getchildren()
                    if not value_els:
                        docstring = None
                    else:
                        first_value_el = value_els[0]
                        docstring = get_docstring_from_value(first_value_el)
    return func_name, docstring

def get_funcs_dets_and_docstring(func_els):
    funcs_dets_and_docstring = []
    for func_el in func_els:
        func_name, docstring = get_func_name_docstring(func_el)
        func_dets_and_docstring = (func_el, func_name, docstring)
        funcs_dets_and_docstring.append(func_dets_and_docstring)
    return funcs_dets_and_docstring

@filt_block_advisor(xpath=FUNC_DEFN_XPATH, warning=True)
def docstring_issues(block_dets, *, repeated_message=False):
    """
    Check over function doc strings. Missing doc string, not enough lines to
    cover params, return etc.
    """
    WRAPPING_NEWLINE_N = 2
    example_docstring = layout(f'''\
        def greet(name, greet_word='Hi'):
            """
            Get a greeting for the supplied person.

            :param str name: person being greeted
            :param str greet_word: the word to start the greeting
            :return: a greeting message to the person
            :rtype: str
            """
            greeting = f"{{greet_word}} {{name}} - how are you?"
            return greeting
        ''', is_code=True)
    func_els = block_dets.element.xpath(FUNC_DEFN_XPATH)
    overall_func_type_lbl = get_overall_func_type_lbl(func_els)
    funcs_dets_and_docstring = get_funcs_dets_and_docstring(func_els)
    brief_msg = ''
    missing_commented = False
    inadequate_commented = False
    for func_el, func_name, docstring in funcs_dets_and_docstring:
        func_type_lbl = get_func_type_lbl(func_el)
        if docstring is None:
            if not (missing_commented or repeated_message):
                brief_msg += (
                    layout(f"""\
                        ### {overall_func_type_lbl.title()} missing doc string

                        `{func_name}` lacks a doc string - you should probably
                        add one.

                        Note - # comments at the top of the {func_type_lbl} do
                        not work as doc strings. Python completely ignores them.
                        If you add a proper doc string, however, it can be
                        accessed by running help({func_name}) or
                        {func_name}.\_\_doc\_\_. Which is useful when using this
                        {func_type_lbl} in bigger projects e.g. in an IDE
                        (Integrated Development Environment).

                        Here is an example doc string for a simple function
                        using one of several valid formats:

                        """)
                    +
                    example_docstring
                )
                missing_commented = True
            else:
                brief_msg += layout(f"""\

                    `{func_name}` lacks a doc string - you should probably add
                    one.
                    """)
        else:
            n_args = get_n_args(func_el)
            n_doc_lines = len(docstring.split('\n')) - WRAPPING_NEWLINE_N
            too_short = n_doc_lines < (conf.MIN_BRIEF_DOCSTRING + n_args)
            param_str = ' given the number of parameters' if n_args > 1 else ''
            if too_short:
                if not (inadequate_commented or repeated_message):
                    brief_msg += (
                        layout(f"""\
                            ### Function doc string too brief?

                            The doc string for {func_name} seems a little
                            short{param_str}. You might want to rework it. Here
                            is an example using one of several valid formats:

                            """)
                        +
                        example_docstring
                    )
                    inadequate_commented = True
                else:
                    brief_msg += layout(f"""\
                        The doc string for {func_name} seems a little short.
                        """)
    if not (missing_commented or inadequate_commented):
        return None
    message = {
        conf.BRIEF: brief_msg,
    }
    return message



import ast
from collections import namedtuple
from functools import partial
import logging

from lxml import etree
import astpath  # @UnresolvedImport

from astpath.asts import _set_encoded_literal, _strip_docstring

## Monkey-patch as at astpath Python 3.8 as at 2020-04-26
## Need to be able to tell val = 1 from val = '1' (that little detail ;-))
def convert_to_xml(node, omit_docstrings=False, node_mappings=None):
    """Convert supplied AST node to XML."""
    possible_docstring = isinstance(node, (ast.FunctionDef, ast.ClassDef, ast.Module))

    xml_node = etree.Element(node.__class__.__name__)
    for attr in ('lineno', 'col_offset'):
        value = getattr(node, attr, None)
        if value is not None:
            _set_encoded_literal(
                partial(xml_node.set, attr),
                value
            )
    if node_mappings is not None:
        node_mappings[xml_node] = node

    node_fields = zip(
        node._fields,
        (getattr(node, attr) for attr in node._fields)
    )

    for field_name, field_value in node_fields:
        if isinstance(field_value, ast.AST):
            field = etree.SubElement(xml_node, field_name)
            field.append(
                convert_to_xml(
                    field_value,
                    omit_docstrings,
                    node_mappings,
                )
            )

        elif isinstance(field_value, list):
            field = etree.SubElement(xml_node, field_name)
            if possible_docstring and omit_docstrings and field_name == 'body':
                field_value = _strip_docstring(field_value)

            for item in field_value:
                if isinstance(item, ast.AST):
                    field.append(
                        convert_to_xml(
                            item,
                            omit_docstrings,
                            node_mappings,
                        )
                    )
                else:
                    subfield = etree.SubElement(field, 'item')
                    _set_encoded_literal(
                        partial(setattr, subfield, 'text'),
                        item
                    )

        elif field_value is not None:

            ## The only change is this immediate function call below
            ## add type attribute e.g. so we can distinguish strings from numbers etc
            ## <Constant lineno="1" col_offset="6" type="int" value="1"/>
            _set_encoded_literal(
                partial(xml_node.set, 'type'),
                type(field_value).__name__
            )

            _set_encoded_literal(
                partial(xml_node.set, field_name),
                field_value
            )

    return xml_node

astpath.asts.convert_to_xml = convert_to_xml

## importing from superhelp only works properly after I've installed superhelp as a pip package (albeit as a link to this code using python3 -m pip install --user -e <path_to_proj_folder>)
## Using this as a library etc works with . instead of superhelp but I want to be be able to run the helper module from within my IDE
from superhelp import advisors, ast_funcs, conf  # @UnresolvedImport
from superhelp.utils import layout_comment  # @UnresolvedImport

BlockDets = namedtuple(
    'BlockDets', 'element, pre_block_code_str, block_code_str, first_line_no')
BlockDets.pre_block_code_str.__doc__ = ("The code up until the line we are "
    "interested in needs to be run - it may depend on names from earlier")

MessageDets = namedtuple('MessageDets',
    'code_str, message, first_line_no, warning, source')
MessageDets.__doc__ += (
    "All the bits and pieces that might be needed to craft a message")
MessageDets.source.__doc__ = ("A unique identifier of the source of message "
    "- useful for auditing / testing")

def get_tree(snippet):
    try:
        tree = ast.parse(snippet)
    except SyntaxError as e:
        raise SyntaxError(
            f"Oops - something seems broken in the snippet - details: {e}")
    return tree

def get_blocks_dets(xml, snippet_lines):
    """
    Returning a list of all the details needed to process a line
    (namely BlockDets named tuples)

    Note - lines in the XML sit immediately under body.

    :return: list of BlockDets named tuples
    :rtype: list
    """
    blocks_dets = []
    all_elements = xml.xpath('body')[0].getchildren()  ## [0] because there is only one body under root
    for element in all_elements:
        first_line_no, last_line_no, _el_lines_n = ast_funcs.get_el_lines_dets(
            element)
        block_code_str = (
            '\n'.join(snippet_lines[first_line_no - 1: last_line_no]).strip())
        pre_block_code_str = (
            '\n'.join(snippet_lines[0: first_line_no - 1]).strip()
            + '\n')
        blocks_dets.append(BlockDets(
            element, pre_block_code_str, block_code_str, first_line_no))
    return blocks_dets

def complete_message(message, *, source):
    """
    Ensure a complete message with all levels is available.

    :param dict message: a dict with as many as three levels
     e.g. conf.BRIEF, MAIN, EXTRA.
    :param str source: usually advisor name but might be the system sending the
     message e.g. if no advice given or a bug happened.
    :return: fixed version of original dict with all levels populated
    :rtype: dict
    """
    try:
        message_keys = message.keys()
    except TypeError:
        raise TypeError(f"message had no keys - message: {message} "
            f"(type {type(message).__name__})")
    if conf.BRIEF not in message_keys:
        raise Exception(f"'{source}' gave a message lacking the mandatory "
            f"'{conf.BRIEF}' level")
    if conf.MAIN not in message_keys:
        message[conf.MAIN] = message[conf.BRIEF]
    if conf.EXTRA not in message_keys:
        message[conf.EXTRA] = ''
    return message

def get_message_dets_from_input(advisor_dets, *,
        advisor_input, code_str, first_line_no, repeated_message):
    name = advisor_dets.advisor_name
    docstring = advisor_dets.advisor.__doc__
    if not docstring:
        raise Exception(f'Advisor "{name}" lacks a docstring - add one!')
    try:
        try:
            message = advisor_dets.advisor(advisor_input,
                repeated_message=repeated_message)
        except TypeError as e:
            if "unexpected keyword argument 'repeated_message'" in str(e):
                message = advisor_dets.advisor(advisor_input)
            else:
                raise
    except Exception as e:
        message = {
            conf.BRIEF: (
                layout_comment(f"""\
                    ### Advisor "`{name}`" unable to run

                    Advisor {name} unable to run. Advisor description:
                    """)
                +  ## show first line of docstring (subsequent lines might have more technical, internally-oriented comments)
                layout_comment(docstring.lstrip('\n').split('\n\n')[0] + '\n')
                +
                layout_comment(str(e))
            )
        }
        source = conf.SYSTEM_MESSAGE
        warning = True
    else:
        source = name
        warning = advisor_dets.warning
        if message is None:
            return None
    message = complete_message(message, source=source)
    message_dets = MessageDets(
        code_str, message, first_line_no, warning, source=name)
    return message_dets

def _get_ancestor_block_element(element):
    """
    The item immediately under body that this element is under.
    """
    ancestor_elements = element.xpath('ancestor-or-self::*')
    ancestor_block_element = ancestor_elements[2]  ## [0] will be Module, 1 is body, and blocks are the children of body
    return ancestor_block_element

def _get_filtered_blocks_dets(advisor_dets, xml, blocks_dets):
    """
    Identify source block elements according to xpath supplied. Then filter
    blocks_dets accordingly.

    :return: filtered_blocks_dets
    :rtype: list
    """
    matching_elements = xml.xpath(advisor_dets.xpath)
    if matching_elements:
        logging.debug(f"{advisor_dets.advisor_name} had at least one match")
    else:
        logging.debug(f"{advisor_dets.advisor_name} had no matches")
    filt_block_elements = set(
        [_get_ancestor_block_element(element) for element in matching_elements])
    filtered_blocks_dets = [block_dets for block_dets in blocks_dets
        if block_dets.element in filt_block_elements]
    if filtered_blocks_dets:
        logging.debug(f"{advisor_dets.advisor_name} had at least one block")
    else:
        logging.debug(f"{advisor_dets.advisor_name} had no blocks")
    return filtered_blocks_dets

def get_block_level_messages_dets(blocks_dets, xml):
    """
    For each advisor, get advice on every relevant block. Element type specific
    advisors process filtered blocks_dets; all block advisors process all blocks
    (as you'd expect ;-)).

    As we iterate through the blocks, only the first block under an advisor
    should get the full message.
    """
    messages_dets = []
    all_advisors_dets = (
        advisors.FILT_BLOCK_ADVISORS + advisors.ANY_BLOCK_ADVISORS)
    for advisor_dets in all_advisors_dets:
        logging.debug(f"About to process '{advisor_dets.advisor_name}'")
        element_filtering = hasattr(advisor_dets, 'xpath')
        if element_filtering:
            filtered_blocks_dets = _get_filtered_blocks_dets(
                advisor_dets, xml, blocks_dets)
            blocks_dets2use = filtered_blocks_dets
            logging.debug(
                f"'{advisor_dets.advisor_name}' has element filtering for "
                f"{len(blocks_dets2use)} matching blocks")
        else:  ## no filtering by element type so process all blocks
            blocks_dets2use = blocks_dets
        repeated_message = False
        for block_dets in blocks_dets2use:
            message_dets = get_message_dets_from_input(advisor_dets,
                advisor_input=block_dets, code_str=block_dets.block_code_str,
                first_line_no=block_dets.first_line_no,
                repeated_message=repeated_message)
            if message_dets:
                messages_dets.append(message_dets)
                repeated_message = True
    return messages_dets

def get_overall_snippet_messages_dets(snippet, blocks_dets):
    """
    Returns messages which apply to snippet as a whole, not just specific
    blocks. E.g. looking at every block to look for opportunities to unpack.
    """
    messages_dets = []
    for advisor_dets in advisors.SNIPPET_ADVISORS:
        message_dets = get_message_dets_from_input(advisor_dets,
            advisor_input=blocks_dets, code_str=snippet, first_line_no=None,
            repeated_message=False)
        if message_dets:
            messages_dets.append(message_dets)
    return messages_dets

def get_separated_messages_dets(snippet):
    """
    Break snippet up into syntactical parts and blocks of code. Apply advisor
    functions and get message details. Split into overall messages and block-
    specific messages.

    :param str snippet: code snippet
    :return: a tuple of two MessageDets lists
     (overall_snippet_messages_dets, block_level_messages_dets)
     or None if no messages
    :rtype: tuple (or None)
    """
    tree = get_tree(snippet)
    xml = astpath.asts.convert_to_xml(tree)
    if conf.RECORD_AST:
        xml.getroottree().write(str(conf.AST_OUTPUT_XML), pretty_print=True)
        logging.info("""\



        Updating AST



        """)
    snippet_lines = snippet.split('\n')
    blocks_dets = get_blocks_dets(xml, snippet_lines)
    overall_snippet_messages_dets = get_overall_snippet_messages_dets(
        snippet, blocks_dets)
    block_level_messages_dets = get_block_level_messages_dets(blocks_dets, xml)
    for messages_dets in [
            overall_snippet_messages_dets, block_level_messages_dets]:
        if None in messages_dets:
            raise Exception("messages_dets in meant to be a list of "
                "MessageDets named tuples yet a None item was found")
    if not (overall_snippet_messages_dets or block_level_messages_dets):
        message = {
            conf.BRIEF: conf.NO_ADVICE_MESSAGE,
        }
        message = complete_message(message, source=conf.SYSTEM_MESSAGE)
        overall_snippet_messages_dets = [
            MessageDets(snippet, message,
                first_line_no=None, warning=False, source=conf.SYSTEM_MESSAGE)]
    return overall_snippet_messages_dets, block_level_messages_dets

def get_error_messages_dets(e, snippet):
    """
    If unable to produce any messages, supply the problem in the form of
    standard messages_dets so the displayers can operate in their usual
    messages_dets consuming ways :-).
    """
    message = {
        conf.BRIEF: layout_comment(f"""\
            ### No advice sorry :-(

            Unable to provide advice - some sort of problem.

            Details: {e}
            """),
    }
    message = complete_message(message, source=conf.SYSTEM_MESSAGE)
    overall_messages_dets = [
        MessageDets(snippet, message,
            first_line_no=None, warning=True, source=conf.SYSTEM_MESSAGE)]
    block_messages_dets = []
    messages_dets = (overall_messages_dets, block_messages_dets)
    return messages_dets



import logging
from textwrap import dedent

from .cli_extras import md2cli

"""
Note - displays properly in the terminal but not necessarily in other output
e.g. Eclipse console
"""

from .. import conf

TERMINAL_WIDTH = 220

SHORT_LINE = '-' * 2  ## at 3 long it automatically becomes a long line (at least, it does in my bash terminal on Ubuntu)
LONG_LINE = '-' * 120
MDV_CODE_BOUNDARY = "```"

def get_message(message_dets, message_level):
    message = dedent(message_dets.message[message_level])
    if message_level == conf.EXTRA:
        message = dedent(message_dets.message[conf.MAIN]) + message
    warning_str = 'WARNING:\n' if message_dets.warning else ''
    message = dedent(warning_str + message)
    message = (message
        .replace(f"    {conf.PYTHON_CODE_START}", MDV_CODE_BOUNDARY)
        .replace(f"\n    {conf.PYTHON_CODE_END}", MDV_CODE_BOUNDARY)
    )
    message = md2cli.main(md=message.replace('`', ''))  ## They create problems in formatting
    return message

def display(snippet, messages_dets, *,
        message_level=conf.BRIEF, in_notebook=False):
    """
    Show by code blocks.
    """
    logging.debug(f"{__name__} doesn't use in_notebook setting {in_notebook}")
    md2cli.term_columns = TERMINAL_WIDTH
    text = [md2cli.main(f"{LONG_LINE}\n"
        "# SuperHELP - Help for Humans!\n"),
        md2cli.main(f"\n{SHORT_LINE}\n"),
        "Help is provided for your overall snippet "
        "and for each line as appropriate.\n",
        f"Currently showing {message_level} content as requested",
        conf.MISSING_ADVICE_MESSAGE,
    ]
    text.append(md2cli.main(dedent(
        "## Overall Snippet\n"
        f"{MDV_CODE_BOUNDARY}\n"
        + snippet
        + f"\n{MDV_CODE_BOUNDARY}")))
    overall_messages_dets, block_messages_dets = messages_dets
    for message_dets in overall_messages_dets:
        message = get_message(message_dets, message_level)
        text.append(message)
    block_messages_dets.sort(key=lambda nt: (nt.first_line_no, nt.warning))
    prev_line_no = None
    for n, message_dets in enumerate(block_messages_dets):
        if n > 41:
            has_prob = True
            print(has_prob)
        ## display code for line number (once ;-))
        line_no = message_dets.first_line_no
        new_block = (line_no != prev_line_no)
        if new_block:
            block_has_warning_header = False
            text.append(md2cli.main(
                f'{LONG_LINE}\n## Code block starting line {line_no:,}'))
            text.append(md2cli.main(dedent(
                f"{MDV_CODE_BOUNDARY}\n"
                + message_dets.code_str
                + f"\n{MDV_CODE_BOUNDARY}")))
            prev_line_no = line_no
        if message_dets.warning and not block_has_warning_header:
            text.append("\n### Questions / Warnings")
            text.append("\nThere may be some issues with this code "
                "block you want to address.")
            block_has_warning_header = True
        ## process message
        message = get_message(message_dets, message_level)
        text.append(message)
    content = '\n'.join(text)
    print(content)



from ..advisors import filt_block_advisor
from ..ast_funcs import get_assign_name
from .. import code_execution, conf, utils
from ..utils import layout_comment as layout

ASSIGN_FUNC_NAME_XPATH = 'descendant-or-self::Assign/value/Call/func/Name'

@filt_block_advisor(xpath=ASSIGN_FUNC_NAME_XPATH)
def set_overview(block_dets, *, repeated_message=False):
    """
    Look for sets and provide general advice on using them and finding out more.
    """
    func_type_els = block_dets.element.xpath(ASSIGN_FUNC_NAME_XPATH)
    has_set = False
    name_sets = []
    for func_type_el in func_type_els:
        try:
            is_set = func_type_el.get('id') == 'set'
        except IndexError:
            is_set = False        
        if not is_set:
            continue
        has_set = True
        name = get_assign_name(func_type_el)
        my_set = code_execution.get_val(
            block_dets.pre_block_code_str, block_dets.block_code_str, name)
        name_sets.append((name, my_set))
    if not has_set:
        return None
    brief_msg = ''
    no_duplicates_demo = ''
    for name, my_set in name_sets:
        empty_set = len(my_set) == 0
        if not empty_set:
            members = str(sorted(my_set)).strip('[').strip(']')
            brief_msg += layout(f"""\
                `{name}` is a set with {utils.int2nice(len(my_set))} members:
                {members}
                """)
            set_item = list(my_set)[0]
            if not no_duplicates_demo and not repeated_message:  ## only want one demo even if multiple sets defined in block
                no_duplicates_demo = (
                    layout(f"""\

                    For example:

                    """)
                    +
                    layout(f"""\
                        {name}.add({set_item})
                        ## >>> {my_set}
                        """, is_code=True)
                )
        else:
            brief_msg += layout(f"""\
                `{name}` is an empty set.
                """)
    if not repeated_message:
        brief_msg += (
            layout("""\

                Python sets are brilliant. There are often cases in programming
                where you need some sort of set operation e.g. you need
                everything in one set that is not in another. In Python you can
                express that idea directly and semantically with set concepts
                instead of having to build the operations yourself in code which
                needs explaining and testing.

                For example:
            """)
            +
            layout("""\
                people = set(['Sam', 'Avi', 'Terri', 'Noor', 'Hyeji'])
                no_email = set(['Sam', 'Terri'])
                people2email = people - no_email
                ## >>> {'Noor', 'Hyeji', 'Avi'}
                """, is_code=True)
        )
    if repeated_message:
        main_msg = brief_msg
        extra_msg = ''
    else:
        main_msg = (
            brief_msg
            +
            layout(f"""\

                Being a set, all members are unique by definition
                so if you add something to a set that is already a member
                the set doesn't change.
                """)
            +
            no_duplicates_demo
            +
            layout("""\

                This is unlike a list which grows each time you append an item
                - the list simply repeats the item. For example:
                """)
            +
            layout("""\
                my_list = [1, 2, 3]
                my_list.append(4)
                my_list.append(4)
                my_list.append(4)
                ## >>> [1, 2, 3, 4, 4, 4]
                my_set = {1, 2, 3}
                my_set.add(4)
                my_set.add(4)
                my_set.add(4)
                ## >>> {1, 2, 3, 4}
                """, is_code=True)
        )
        extra_msg = (
            layout("""\
                Set operations can be expressed with operators such as
                '-' (minus) or with methods such as .difference().

                Sets are well explained in the official documentation so it will
                suffice to show a simple example.

                Note - the operator corresponding to the .union() method is the pipe
                '|' not '+'.
                """)
            +
            layout("""\
                badminton_players = set(['Grant', 'Charlotte', 'Aravind'])
                tennis_players = set(['Giles', 'Grant'])
                squash_players = set(['Grzegorz'])
                racquet_players = badminton_players | tennis_players | squash_players
                ## >>> {'Grzegorz', 'Giles', 'Grant', 'Aravind', 'Charlotte'}
                """, is_code=True)
        )
    message = {
        conf.BRIEF: brief_msg,
        conf.MAIN: main_msg,
        conf.EXTRA: extra_msg,
    }
    return message



from ..advisors import shared, filt_block_advisor
from ..ast_funcs import get_el_lines_dets
from .. import conf
from ..utils import layout_comment as layout

FOR_XPATH = 'descendant-or-self::For'

@filt_block_advisor(xpath=FOR_XPATH)
def comprehension_option(block_dets, *, repeated_message=False):
    """
    Provide overview of for loop to see if simple enough to be a possible
    candidate for a comprehension.

    Don't try to properly understand the for loop or make a built comprehension.
    It is enough to detect whether a loop is simple enough to consider making a
    comprehension or not. And to see whether appending, key setting, or adding
    is happening and suggesting the right comprehension accordingly.
    """
    for_els = block_dets.element.xpath(FOR_XPATH)
    any_short_enough = False
    for for_el in for_els:
        _first_line_no, _last_line_no, for_lines_n = get_el_lines_dets(
            for_el, ignore_trailing_lines=True)
        short_enough = for_lines_n < 3
        if short_enough:
            any_short_enough = True
            break
    if not any_short_enough:
        return None
    comp_type = None
    comp_comment = ''
    if 'append' in block_dets.block_code_str:
        comp_type = 'List Comprehension'
        comp_comment = shared.LIST_COMPREHENSION_COMMENT
    elif len(block_dets.element.cssselect('Subscript')):  ## Seems a reasonable indicator
        comp_type = 'Dictionary Comprehension'
        comp_comment = shared.DICT_COMPREHENSION_COMMENT
    elif 'set' in block_dets.block_code_str:
        comp_type = 'Set Comprehension'
        comp_comment = shared.SET_COMPREHENSION_COMMENT
    else:
        return None
    brief_msg = layout(f"""\
        ### Possible option of using a {comp_type}

        """)
    if not repeated_message:
        brief_msg += layout(f"""\
            Simple for loops can sometimes be replaced with comprehensions. In
            this case a simple reading of the code suggests a {comp_type} might
            be possible. Of course, only use a comprehension if it makes your
            code easier to understand.
            """)
    message = {
        conf.BRIEF: brief_msg,
        conf.MAIN: (
            brief_msg
            +
            shared.GENERAL_COMPREHENSION_COMMENT + '\n\n' + comp_comment
        ),
    }
    return message

def get_incremental_iteration_dets(for_el):
    """
    Look for the pattern:

    pets = ['cat', 'dog']
    for i in range(len(pets)):
        print(f"My {pets[i]}")

    For/target/Name id = i
        iter/Call/func/Name id = range
                  args/Call/func/Name id = len
                            args/Name id = pets

    ... Subscript/value/Name id = pets
                  slice/Index/value/Name = i
    """
    target_name_els = for_el.xpath('target/Name')
    if not target_name_els:
        return None
    index_name = target_name_els[0].get('id')  ## e.g. 'i'
    iterator_func_name_els = for_el.xpath('iter/Call/func/Name')
    if not iterator_func_name_els:
        return None
    range_iterator_func = iterator_func_name_els[0].get('id') == 'range'  ## range()
    if not range_iterator_func:
        return None
    iter_arg_func_name_els = for_el.xpath('iter/Call/args/Call/func/Name')
    if not iter_arg_func_name_els:
        return None
    len_inside = (iter_arg_func_name_els[0].get('id') == 'len')  ## range(len())
    if not len_inside:
        return None
    len_arg_name_els = for_el.xpath('iter/Call/args/Call/args/Name')
    if not len_arg_name_els:
        return None
    iterable_name = len_arg_name_els[0].get('id')  ## e.g. pets
    if not iterable_name:
        return None
    subscript_name_els = for_el.xpath(
        'descendant-or-self::Subscript/value/Name')
    if not subscript_name_els:
        return None
    subscript_name = subscript_name_els[0].get('id')  ## e.g. pets
    slicing_iterable = (iterable_name == subscript_name)
    if not slicing_iterable:
        return None
    slice_index_name_els = for_el.xpath(
        'descendant-or-self::Subscript/slice/Index/value/Name')
    if not slice_index_name_els:
        return None
    slice_index_name = slice_index_name_els[0].get('id')
    slicing_by_range_val = (slice_index_name == index_name)
    if not slicing_by_range_val:
        return None
    return index_name, iterable_name

@filt_block_advisor(xpath=FOR_XPATH)
def for_index_iteration(block_dets, *, repeated_message=False):
    """
    Look to see if an opportunity for simple iteration available as more
    pythonic alternative to incremental indexing.

    Look for the pattern:

    for i in range(len(foo)):
        foo[i] detected
    """
    for_els = block_dets.element.xpath(FOR_XPATH)
    any_incremental_iteration = False
    for for_el in for_els:
        try:
            index_name, iterable_name = get_incremental_iteration_dets(for_el)
        except TypeError:
            continue
        else:
            any_incremental_iteration = True
            break
    if not any_incremental_iteration:
        return None
    brief_msg = layout(f"""\

        ### Possible option of using direct iteration

        It looks like your snippet iterates through `{iterable_name}` using
        indexes. In Python you can iterate directly which is much easier.
        """)
    if not repeated_message:
        brief_msg += (
            layout(f"""\

                For example, instead of:

                """)
            +
            layout(f"""\

                for {index_name} in range(len({iterable_name})):
                    print({iterable_name}[{index_name}])

                """, is_code=True)
            +
            layout(f"""\

                you can directly iterate as follows:

                """)
            +
            layout(f"""\

                for item in {iterable_name}:  ## item should be replaced with a more useful name
                    print(item)

                """, is_code=True)
            +
            layout(f"""\

                which is considered more pythonic i.e. good.

                """)
        )
    message = {
        conf.BRIEF: brief_msg,
    }
    return message

@filt_block_advisor(xpath=FOR_XPATH)
def nested_fors(block_dets, *, repeated_message=False):
    """
    Look to see if an opportunity for using itertools.product instead of nested
    iteration.
    """
    for_els = block_dets.element.xpath(FOR_XPATH)
    nested_iteration = False
    for for_el in for_els:
        nested_for_els = for_el.xpath('descendant::For')
        if nested_for_els:
            nested_iteration = True
            break
    if not nested_iteration:
        return None
    brief_msg = layout("""\

        ### Possible option of simplifying nested iteration

        Consider replacing nested iteration with `itertools.product`.

        """)
    if not repeated_message:
        brief_msg += (
            layout("""\
                For example, you could replace:

                """)
            +
            layout("""\
                for person in persons:
                    for pet in pets:
                        for year in years:
                            print(f"{person} might like a {pet} in {year}")
                """, is_code=True)
            +
            layout("""\

                with a version using `itertools.product`:

                """)
            +
            layout("""\
                from itertools import product
                for person, pet, year in product(persons, pets, years):
                    print(f"{person} might like a {pet} in {year}")
                """, is_code=True)
        )
    main_msg = brief_msg
    if not repeated_message:
        main_msg += layout("""\

            Whether this is a good idea or not depends on your specific code but
            using `product` has the advantage of reducing indentation. It also
            semantically expresses the intention of the code - namely to look at
            every option in the cartesian product of the different collections
            of items.
            """)
    message = {
        conf.BRIEF: brief_msg,
        conf.MAIN: main_msg,
    }
    return message



from ..advisors import snippet_advisor
from .. import conf
from ..utils import get_nice_str_list, int2nice, layout_comment as layout

UNSPECIFIC_EXCEPTION = 'Exception'

def _get_exception_block_comment(exception_block):
    handlers = get_nice_str_list(exception_block, quoter='`')
    comment = f"The following exception handlers were detected: {handlers}"
    return comment

def get_exception_blocks(blocks_dets):
    """
    There can be multiple try-except statements in a snippet so we have to
    handle each of them.
    """
    exception_blocks = []
    for block_dets in blocks_dets:
        block_exception_types = []
        except_handler_els = block_dets.element.xpath(
            'descendant-or-self::handlers/ExceptHandler')
        for except_handler_el in except_handler_els:
            exception_type_els = except_handler_el.xpath(
                'type/Name | type/Tuple/elts/Name')
            exception_handler_types = [
                exception_type_el.get('id')
                for exception_type_el in exception_type_els]
            block_exception_types.extend(exception_handler_types)
        if block_exception_types:
            exception_blocks.append(block_exception_types)
    return exception_blocks

@snippet_advisor()
def exception_overview(blocks_dets):
    """
    Provide overview of exception handling.
    """
    exception_blocks = get_exception_blocks(blocks_dets)
    if not exception_blocks:
        return None
    brief_msg = '### Exception handling'
    for n, exception_block in enumerate(exception_blocks, 1):
        counter = '' if len(exception_blocks) == 1 else f" {int2nice(n)}"
        brief_msg += layout(f"""\

            #### `try`-`except` block{counter}

            """)
        brief_msg += _get_exception_block_comment(exception_block)
    message = {
        conf.BRIEF: layout(brief_msg),
    }
    return message

@snippet_advisor(warning=True)
def unspecific_exception(blocks_dets):
    """
    Look for unspecific exceptions.
    """
    exception_blocks = get_exception_blocks(blocks_dets)
    if not exception_blocks:
        return None
    brief_msg = ''
    unspecific_block_ns = []
    has_title = False
    has_unspecific = False
    for n, exception_block in enumerate(exception_blocks, 1):
        only_unspecific = (len(exception_block) == 1
            and exception_block[0] == UNSPECIFIC_EXCEPTION)
        if not only_unspecific:
            continue
        has_unspecific = True
        unspecific_block_ns.append(n)
        if not has_title:
            brief_msg += layout("""\

                #### Un-specific `Exception` only in `try`-`except` block(s)

                """)
        has_title = True
    if not has_unspecific:
        return None
    n_unspecific = len(unspecific_block_ns)
    if n_unspecific == 1:
        block_n_specific_text = ' has'
    else:
        unspecific_nice_block_ns = [
            int2nice(unspecific_block_n)
            for unspecific_block_n in unspecific_block_ns]
        blocks_ns = get_nice_str_list(unspecific_nice_block_ns, quoter='')
        block_n_specific_text = f"s {blocks_ns} have"
    brief_msg += layout(f"""\

        `try`-`except` block{block_n_specific_text} an un-specific Exception
        only.

        Using the un-specific exception type `Exception` is often completely
        appropriate. But if you are looking for specific exceptions you
        should handle those separately.

        For example:
        """)
    message = {
        conf.BRIEF: layout(brief_msg),
        conf.MAIN: (
            layout(brief_msg)
            +
            layout("""\

                try:
                    spec_dicts[idx][spec_type]
                except IndexError:
                    print(f"Unable to access requested spec_dict (idx {idx})")
                except KeyError:
                    print(f"Unable to access '{spec_type}' for "
                        f"requested spec_dict (idx {idx})")
                except Exception as e:
                    print(f"Unexpected exception - details: {e}")
                """, is_code=True)
        ),
    }
    return message



from collections import defaultdict

from ..advisors import any_block_advisor
from .. import conf, utils
from .shared import is_reserved_name
from ..utils import get_nice_str_list, int2nice, layout_comment as layout

def _get_shamed_names_title(reserved_names, bad_names, dubious_names):
    if not (reserved_names or bad_names or dubious_names):
        raise Exception(f"No shamed names supplied to {__name__}")
    n_reserved_names = len(reserved_names)
    n_bad_names = len(bad_names)
    n_dubious_names = len(dubious_names)
    if reserved_names:
        if n_reserved_names == 1:
            title = "Bad naming - use of reserved name."
        else:
            title = "Bad naming - use of reserved names."
    else:
        title = ''
    if title:
        if bad_names:
            if n_bad_names == 1:
                if n_dubious_names < 1:
                    title += ' Also un-pythonic name'
                else:
                    title += ' Also un-pythonic name(s)'
            else:
                title += ' Also un-pythonic names'
        elif dubious_names:
            if n_dubious_names == 1:
                title += ' Also a possibly un-pythonic name'
            else:
                title += ' Also other possibly un-pythonic names'
    else:
        if bad_names:
            if n_bad_names == 1:
                if n_dubious_names < 1:
                    title = 'Un-pythonic name'
                else:
                    title = 'Un-pythonic name(s)'
            else:
                title = 'Un-pythonic names'
        elif dubious_names:
            if n_dubious_names == 1:
                title = 'Possibly an un-pythonic name'
            else:
                title = 'Possibly some un-pythonic names'
    return title

def _get_shamed_names_comment(shamed_names):
    multiple_shamed_names = len(shamed_names) > 1
    if multiple_shamed_names:
        shamed_names_listed = utils.get_nice_str_list(shamed_names, quoter='`')
        shamed_names_comment = f"{shamed_names_listed} are un-pythonic."
    else:
        shamed_names_comment = f"`{shamed_names[0]}` is un-pythonic."
    return shamed_names_comment

def get_standard_assigned_names(block_dets):
    """
    Only get names where we expect standard pythonic naming. So not named tuple
    or class names, for example. We have to explicitly ignore named tuple names.
    Classes are automatically excluded because I haven't explicitly included
    them. They are stored differently e.g. <ClassDef ...name="ActuallyGoodName">
    """
    assigned_name_els = block_dets.element.xpath(
        'descendant-or-self::targets/Name | descendant-or-self::target/Name')
    assigned_names = []
    for name_el in assigned_name_els:
        ## exclude if named tuple - they are allowed "un-Pythonic" names
        try:
            assign_el = name_el.xpath('ancestor::Assign')[0]
        except IndexError:
            ## cannot be a named tuple
            pass
        else:
            func_name_els = assign_el.xpath('value/Call/func/Name')
            if func_name_els:
                func_names = [
                    func_name_el.get('id') for func_name_el in func_name_els]
                if 'namedtuple' in func_names:
                    continue
        ## not a named tuple 
        name = name_el.get('id')
        assigned_names.append(name)
    return assigned_names

def get_class_names(block_dets):
    class_els = block_dets.element.xpath('ClassDef')
    class_names = [class_el.get('name') for class_el in class_els]
    return class_names

def get_named_tuple_names(block_dets):
    assigned_name_els = block_dets.element.xpath(
        'descendant-or-self::Assign/targets/Name')
    named_tuple_names = []
    for name_el in assigned_name_els:
        assign_el = name_el.xpath('ancestor::Assign')[0]
        func_name_els = assign_el.xpath('value/Call/func/Name')
        if func_name_els:
            func_names = [
                func_name_el.get('id') for func_name_el in func_name_els]
            if 'namedtuple' not in func_names:
                continue
        name = name_el.get('id')
        named_tuple_names.append(name)
    return named_tuple_names

def get_unpacked_names(block_dets):
    unpacked_name_els = block_dets.element.xpath(
        'descendant-or-self::targets/Tuple/elts/Name'
        ' | '
        'descendant-or-self::target/Tuple/elts/Name')
    unpacked_names = [unpacked_name_el.get('id')
        for unpacked_name_el in unpacked_name_els]
    return unpacked_names

def get_def_func_names(block_dets):
    def_func_elements = block_dets.element.xpath(
        'descendant-or-self::FunctionDef')
    def_func_names = [name_el.get('name') for name_el in def_func_elements]
    return def_func_names

def get_all_names(block_dets, *, include_non_standard=False):
    """
    :param bool include_non_standard: if True include variables that aren't
     expected to follow standard Python naming conventions e.g. class or named
     tuple names.
    """
    assigned_names = get_standard_assigned_names(block_dets)
    unpacked_names = get_unpacked_names(block_dets)
    def_func_names = get_def_func_names(block_dets)
    all_names = assigned_names + unpacked_names + def_func_names
    if include_non_standard:
        class_names = get_class_names(block_dets)
        named_tuple_names = get_named_tuple_names(block_dets)
        all_names = all_names + class_names + named_tuple_names
    return all_names

@any_block_advisor(warning=True)
def unpythonic_name_check(block_dets, *, repeated_message=False):
    """
    Check names used for use of reserved words and camel case.
    """
    names = get_all_names(block_dets, include_non_standard=False)
    reserved_names = []
    bad_names = []
    dubious_names = []
    for name in names:
        if is_reserved_name(name):
            reserved_names.append(name)
        all_lower_case = (name.lower() == name)
        all_upper_case = (name.upper() == name)
        bad_name = not (all_lower_case or all_upper_case)
        if bad_name:
            bad_names.append(name)
        elif all_upper_case:
            dubious_names.append(name)
    if not (reserved_names or bad_names or dubious_names):
        return None
    title = _get_shamed_names_title(reserved_names, bad_names, dubious_names)
    brief_msg = layout(f"""\
        ### {title}

        """)
    if reserved_names:
        reserved_names_listed = utils.get_nice_str_list(
            reserved_names, quoter='`')
        brief_msg += layout(f"""\
            Reserved name(s): {reserved_names_listed}
            """)
    if bad_names:
        bad_names_listed = utils.get_nice_str_list(bad_names, quoter='`')
        brief_msg += layout(f"""\
            Un-pythonic name(s): {bad_names_listed}
            """)
    if dubious_names:
        dubious_names_listed = utils.get_nice_str_list(
            dubious_names, quoter='`')
        brief_msg += layout(f"""\
            Possibly un-pythonic name(s): {dubious_names_listed}
            """)
    if not repeated_message:
        brief_msg += layout("""\

            Python variables should not named using reserved words e.g.
            `collections` or `sorted`.

            Generally speaking Python variables should be snake case - that is
            lower case, with multiple words joined by underscores e.g.
            `high_scores` (not `highScores` or `HighScores`)
            """)
    main_msg = brief_msg
    if not repeated_message:
        main_msg += layout("""\
            In Python class names and named tuples are expected to be in Pascal
            Case (also known as upper camel case) rather than the usual snake
            case. E.g. `collections.ChainMap`

            Exceptions can also be made when a higher priority is consistency
            with other code e.g. a library the Python is ported from, or the
            non-Python code that Python is wrapping.
            """)
    message = {
        conf.BRIEF: brief_msg,
        conf.MAIN: main_msg,
    }
    return message

@any_block_advisor(warning=True)
def short_name_check(block_dets, *, repeated_message=False):
    """
    Check for short variable names.
    """
    names = get_all_names(block_dets, include_non_standard=True)
    short_names = defaultdict(set)
    for name in names:
        if len(name) < conf.MIN_BRIEF_NAME:
            short_names[len(name)].add(name)
    if not short_names:
        return None
    short_comment = ''
    for length, names in sorted(short_names.items()):
        freq = len(names)
        multiple = (freq > 1)
        if multiple:
            nice_list = get_nice_str_list(list(names), quoter='`')
            short_comment += (
                f" {int2nice(freq).title()} variables have names "
                f"{int2nice(length)} characters long: {nice_list}.")
        else:
            name = names.pop()
            short_comment += f"`{name}` is short."
    brief_msg = layout(f"""\
        ### Short variable names

        Sometimes, short variable names are appropriate - even conventional -
        but they should be avoided outside of a few special cases. In your code:

        {short_comment}
        """)
    main_msg = brief_msg
    if not repeated_message:
        main_msg += (
            layout("""\

                In many programming languages it is idiomatic to use `i`, `j`,
                and even `k` as increment counters. It is best to reserve `i`
                for incrementing starting at 0 and prefer `n` when starting at
                1. But longer names should be used when they aid readability
                (i.e. usually ;-)).

               `k` and `v` are idiomatic in Python when iterating through
               dictionary items e.g.

            """)
            +
            layout("""\

                for k, v in my_dict.items():
                    ...

            """, is_code=True)
            +
            layout("""\
                But even they should probably be replaced with something more
                descriptive.

                The main goal is readability, readability, readability. That is
                what should drive variable naming above all else. Only a modest
                weight should be given to speed of typing in any code that is
                going to be used or worked on in the future as opposed to quick
                exploratory code in a terminal / notebook etc.

                In the words of the legendary Donald Knuth:

                > "Programs are meant to be read by humans
                >  and only incidentally for computers to execute."

                Note - excessively long variable names can be unreadable as
                well. They may also indicate the code needs to be reworked.
                """)
        )
    message = {
        conf.BRIEF: brief_msg,
        conf.MAIN: main_msg,
    }
    return message



from html import unescape
import markdown  # @UnresolvedImport
from markdown.extensions import fenced_code  # @UnresolvedImport
from markdown.extensions.tables import TableExtension  # @UnresolvedImport

from . import ansi_printer, cli_utils, tag_formatting

def get_ansi(md):
    """
    Covert markdown string into formatted and code-highlighted terminal code.
    """
    MD = markdown.Markdown(
        tab_length=4,
        extensions=[
            ansi_printer.AnsiPrintExtension(),
            TableExtension(),
            fenced_code.FencedCodeExtension(),
        ],
    )
    MD.convert(md)
    ansi = MD.ansi + '\n'
    ansi = cli_utils.set_hr_widths(ansi) + "\n"
    # The RAW html within source, incl. fenced code blocks:
    # phs are numbered like this in the md, we replace back:
    PH = markdown.util.HTML_PLACEHOLDER
    stash = MD.htmlStash
    for nr, ph in enumerate(stash.rawHtmlBlocks):
        raw = unescape(ph)
        if raw[:3].lower() == '<br':
            raw = '\n'
        pre = '<pre><code'
        if raw.startswith(pre):
            _, raw = raw.split(pre, 1)
            raw = raw.split('>', 1)[1].rsplit('</code>', 1)[0]
            raw = tag_formatting.code(raw.strip(), from_fenced_block=True)
        ansi = ansi.replace(PH % nr, raw)
    return ansi

def main(md):
    ansi = get_ansi(md) if md else ''
    return ansi



from ..advisors import filt_block_advisor
from .. import code_execution, conf, utils
from ..ast_funcs import get_assign_name
from ..utils import get_nice_str_list, layout_comment as layout

ASSIGN_DICT_XPATH = 'descendant-or-self::Assign/value/Dict'

def _get_additional_main_msg(first_name):
    additional_main_msg = (
        layout(f"""

            It is common to iterate through the key-value pairs of a dictionary.
            This can be achieved using the dictionary's `.items()` method. E.g.

            """)
        +
        layout(f"""\
            ## k, v is conventional, and OK in a hurry, but readable names
            ## are probably better for code you're going to maintain
            for k, v in {first_name}.items():
                print(f"key {{k}} maps to value {{v}}")
            """, is_code=True)
        +
        layout(f"""
            
            Keys are unique but values can be repeated. For example:

            """)
        +
        layout(f"""
            country2car = {{'Japan': 'Toyota', 'Sweden': 'Volvo'}}  ## OK - all keys are unique
            country2car = {{'Japan': 'Toyota', 'Japan': 'Honda'}}  ## Oops - the 'Japan' key is repeated

            """, is_code=True)
        +
        layout(f"""

            In which case a better structure might be to have each 'value'
            being a list e.g.

            """)
        +
        layout(f"""
            country2cars = {{'Japan': ['Toyota', 'Honda'], 'Sweden': ['Volvo']}}  ## OK - all keys are unique

            """, is_code=True)
    )
    return additional_main_msg

def _get_minimal_dict_details(block_dets, dict_els, plural):
    brief_msg = ''
    for i, dict_el in enumerate(dict_els):
        first = (i == 0)
        name = get_assign_name(dict_el)
        items = code_execution.get_val(
            block_dets.pre_block_code_str, block_dets.block_code_str, name)
        if first:
            title = layout(f"""\

                ### Dictionar{plural} defined

                """)
            brief_msg += title
        brief_msg += layout(f"""\

            `{name}` is a dictionary with {utils.int2nice(len(items))} items
            (i.e. {utils.int2nice(len(items))} mappings).
            """)
    message = {
        conf.BRIEF: brief_msg,
    }
    return message

def _get_full_dict_details(block_dets, dict_els, plural):
    brief_msg = ''
    main_msg = ''
    first_name = None
    for i, dict_el in enumerate(dict_els):
        first = (i == 0)
        name = get_assign_name(dict_el)
        items = code_execution.get_val(
            block_dets.pre_block_code_str, block_dets.block_code_str, name)
        if first:
            first_name = name
            title = layout(f"""\

                ### Dictionar{plural} defined

                """)
            brief_msg += title
            main_msg += title

            brief_msg += layout("""\

                Dictionaries map keys to values.

                """)
            main_msg += layout("""\

                Dictionaries, along with lists, are the workhorses of Python
                data structures.

                """)
        empty_dict = (len(items) == 0)
        if empty_dict:
            list_desc = layout(f"""\

                `{name}` is an empty dictionary.

                """)
        else:
            plural = '' if len(items) == 1 else 's'
            list_desc = layout(f"""\

                `{name}` is a dictionary with {utils.int2nice(len(items))}
                item{plural} (i.e. {utils.int2nice(len(items))}
                mapping{plural}). In this case, the keys are:
                {list(items.keys())}. We can get the keys using the `.keys()`
                method e.g. `{name}`.`keys()`. The values are
                {list(items.values())}. We can get the values using the
                `.values()` method e.g. `{name}`.`values()`.

                """)
        brief_msg += list_desc
        main_msg += list_desc
   
        brief_msg += layout("""\

            Keys are unique but values can be repeated.

            Dictionaries, along with lists, are the workhorses of Python data
            structures.
            """)
    main_msg += _get_additional_main_msg(first_name)
    message = {
        conf.BRIEF: brief_msg,
        conf.MAIN: main_msg,
        conf.EXTRA: layout("""\

            Python dictionaries (now) keep the order in which items are added.

            They are also super-efficient and fast. The two presentations to
            watch are by living treasure Brandon Rhodes:

            1. The Dictionary Even Mightier -
            <https://www.youtube.com/watch?v=66P5FMkWoVU>
            2. The Mighty Dictionary -
            <https://www.youtube.com/watch?v=oMyy4Sm0uBs>
            """)
    }
    return message

@filt_block_advisor(xpath=ASSIGN_DICT_XPATH)
def dict_overview(block_dets, *, repeated_message=False):
    """
    Look at assigned dictionaries e.g. location = {'country' 'New Zealand',
    'city': 'Auckland'}
    """
    dict_els = block_dets.element.xpath(ASSIGN_DICT_XPATH)
    plural = 'ies' if len(dict_els) > 1 else 'y'
    if repeated_message:
        message = _get_minimal_dict_details(block_dets, dict_els, plural)
    else:
        message = _get_full_dict_details(block_dets, dict_els, plural)
    return message

def get_key_type_names(items):
    key_type_names = sorted(set(
        [type(item).__name__ for item in items]
    ))
    key_type_nice_names = [
        conf.TYPE2NAME.get(key_type, key_type)
        for key_type in key_type_names]
    return key_type_names, key_type_nice_names

@filt_block_advisor(xpath=ASSIGN_DICT_XPATH, warning=True)
def mixed_key_types(block_dets, *, repeated_message=False):
    """
    Warns about dictionaries with mix of string and integer keys.
    """
    dict_els = block_dets.element.xpath(ASSIGN_DICT_XPATH)
    brief_msg = ''
    main_msg = ''
    has_mixed = False
    mixed_names = []
    for i, dict_el in enumerate(dict_els):
        first = (i == 0)
        name = get_assign_name(dict_el)
        items = code_execution.get_val(
            block_dets.pre_block_code_str, block_dets.block_code_str, name)
        key_type_names, _key_type_nice_names = get_key_type_names(items)
        bad_key_type_combo = (
            conf.INT_TYPE in key_type_names and conf.STR_TYPE in key_type_names)
        if not bad_key_type_combo:
            continue
        mixed_names.append(name)
        has_mixed = True
        if first:
            title = layout(f"""\

                ### Mix of integer and string keys in dictionary

                """)
            brief_msg += title
            main_msg += title
    if not has_mixed:
        return None
    multiple = len(mixed_names) > 1
    if multiple:
        nice_str_list = get_nice_str_list(mixed_names, quoter='`')
        mixed_warning = layout(f"""

            {nice_str_list} have keys include a mix of strings and integers -
            which is probably a bad idea.
            """)
    else:
        mixed_warning = layout(f"""

            `{name}`'s keys include both strings and integers which is probably
            a bad idea.
            """)
    brief_msg += mixed_warning
    main_msg += mixed_warning

    if not repeated_message:
        main_msg += layout("""\

            For example, if you have both 1 and "1" as keys in a dictionary
            (which is allowed because they are not the same key) it is very easy
            to get confused and create Hard To Find™ bugs. You _might_ not
            regret it but you probably will ;-).
            """)

    message = {
        conf.BRIEF: brief_msg,
        conf.MAIN: main_msg,
    }
    return message




def get_assign_name(element):
    """
    Get name assignment associated with the element. The element might be the
    value or the target or something but we just want to identify the closest
    Assign ancestor and get its Name.

    If Assign appears more than once in the ancestral chain e.g.
    body-Assign-spam-eggs-Assign-targets-Name then we get a list like this:
    [body-Assign, body-Assign-spam-eggs-Assign] and we want the closest one to
    the element i.e. assign_els[-1]

    Ordered set of nodes, from parent to ancestor?
    https://stackoverflow.com/a/15645846
    """
    assign_els = element.xpath('ancestor::Assign')
    assign_el = assign_els[-1]
    name_els = assign_el.xpath('targets/Name')
    if len(name_els) == 1:
        name = name_els[0].get('id')
    else:
        name = None
    return name

def get_el_lines_dets(el, *, ignore_trailing_lines=False):
    """
    How long is the snippet of code that completely wraps up this element? If
    there are no "trailing lines" it is simple because the AST includes line
    numbers. We just get the min and max line number values. But if there are,
    these trailing lines are not part of the AST (in the same way that comments
    are ignored). So:

    d = {
        1: 1}

    d = {
        1: 1,
    }

    are equivalent in the AST. And there are only lines 1-2 in the AST in both
    cases.

    If there are subsequent AST-relevant lines of code, we can find the first of
    those lines, subtract 1, and treat that as the end of our code. If client
    code is worried about trailing empty lines it can strip them off itself.
    Blank lines (or other non-code lines e.g. comments) won't have any effect on
    the execution of the code.

    They do affect our judgement of code length though e.g. to see if the
    snippet is too long or short so the ignore_trailing_lines option is
    provided.

    If there are no subsequent AST-relevant lines of code, we are in the dark
    about where exactly the actual code string (assuming we only have the AST to
    work from). In which case we simply add enough lines to cover all reasonable
    cases e.g. if different levels of nested parentheses/brackets are on
    separate lines. E.g.

    a = (
            (
                (
                    (
                        (
                            (
                                (
                                 ...
                                )
                            )
                        )
                    )
                )
            )
        )
    """
    SAFE_EXTRA_LINES = 10
    line_no_strs = set(el.xpath('descendant-or-self::*[@lineno]/@lineno'))
    line_nos = [int(line_no_str) for line_no_str in line_no_strs]
    if not line_nos:
        first_line_no, last_line_no, el_lines_n = None, None, None
    else:
        first_line_no = min(line_nos)
        last_ast_line_no = max(line_nos)
        if ignore_trailing_lines:
            last_line_no = last_ast_line_no
        else:
            module_el = el.xpath('//Module')[0]
            all_line_no_strs = set(
                module_el.xpath('descendant::*[@lineno]/@lineno'))
            all_line_nos = [
                int(line_no_str) for line_no_str in all_line_no_strs]
            subsequent_line_nos = [line_no for line_no in all_line_nos
                if line_no > last_ast_line_no]
            if not subsequent_line_nos:
                last_line_no = last_ast_line_no + SAFE_EXTRA_LINES
            else:
                last_line_no = min(subsequent_line_nos) - 1
        el_lines_n = last_line_no - first_line_no + 1
    return first_line_no, last_line_no, el_lines_n



from collections import defaultdict

from ..advisors import filt_block_advisor
from ..ast_funcs import get_assign_name
from .. import code_execution, conf
from ..utils import get_nice_str_list, get_python_version,\
    layout_comment as layout

ASSIGN_VAL_XPATH = 'descendant-or-self::Assign/value'

def get_num_els_3_7(block_el):
    num_els = block_el.xpath('descendant-or-self::Assign/value/Num')
    return num_els

def get_num_els_3_8(block_el):
    val_els = block_el.xpath(ASSIGN_VAL_XPATH)
    num_els = []
    for val_el in val_els:
        constant_els = val_el.xpath('Constant') 
        if len(constant_els) != 1:
            continue
        constant_el = constant_els[0]
        if constant_el.get('type') in ('int', 'float'):
            num_els.append(constant_el)
    return num_els

python_version = get_python_version()
if python_version in (conf.PY3_6, conf.PY3_7):
    get_num_els = get_num_els_3_7
elif python_version == conf.PY3_8:
    get_num_els = get_num_els_3_8
else:
    raise Exception(f"Unexpected Python version {python_version}")

@filt_block_advisor(xpath=ASSIGN_VAL_XPATH)
def num_overview(block_dets, *, repeated_message=False):
    """
    Get general advice about assigned numbers e.g. var = 123
    """
    num_els = get_num_els(block_dets.element)
    val_types = defaultdict(list)
    has_num = False
    type_firsts = {}
    for num_el in num_els:
        name = get_assign_name(num_el)
        val = code_execution.get_val(
            block_dets.pre_block_code_str, block_dets.block_code_str, name)
        val_type = type(val).__name__
        val_types[val_type].append(name)
        if not type_firsts.get(val_type):
            type_firsts[val_type] = val
        has_num = True
    if not has_num:
        return None
    brief_msg = ''
    for val_type, names in val_types.items():
        names_text = get_nice_str_list(names, quoter='`')
        if len(names) == 1:
            names_msg = layout(f"""\

                {names_text} is a number - specific type `{val_type}`.
                """)
        else:
            names_msg = layout(f"""\

                {names_text} are numbers - specific type `{val_type}`.
                """)
        brief_msg += names_msg
    if not repeated_message:
        for val_type, names in val_types.items():
            first_name = names[0]
            val = type_firsts[val_type]
            specific_comment = None
            if val_type == conf.INT_TYPE:
                specific_comment = layout(f"""\

                    Integers are counting numbers and include 0 and negative
                    numbers e.g. -2

                    If you need a float instead of an integer use the float
                    function.

                    e.g. float({first_name}) which returns {float(val)}
                    """)
            elif val_type == conf.FLOAT_TYPE:
                specific_comment = layout(f"""\

                    Floats are used when decimal places are required.

                    If you need an integer instead of a float use the int
                    function.

                    e.g. int({name}) which returns {int(val)}
                    """)
            if specific_comment:
                brief_msg += specific_comment
    message = {
        conf.BRIEF: brief_msg,
    }
    if conf.FLOAT_TYPE in val_types and not repeated_message:
        message[conf.EXTRA] = layout(f"""\
            Floats, or floating point numbers, are stored in computers as binary
            fractions. "Unfortunately, most decimal fractions cannot be
            represented exactly as binary fractions. A consequence is that, in
            general, the decimal floating-point numbers you enter are only
            approximated by the binary floating-point numbers actually stored
            in the machine." For more information, read the rest of
            <https://docs.python.org/3/tutorial/floatingpoint.html>. It is
            really interesting - honest!
            """)
    return message



from ..advisors import shared, filt_block_advisor
from ..ast_funcs import get_assign_name
from .. import code_execution, conf, utils
from ..utils import layout_comment as layout

ASSIGN_LISTCOMP_XPATH = 'descendant-or-self::Assign/value/ListComp'

@filt_block_advisor(xpath=ASSIGN_LISTCOMP_XPATH)
def listcomp_overview(block_dets, *, repeated_message=False):
    """
    Provide advice on list comprehensions and explain other types of
    comprehension available in Python.
    """
    listcomp_els = block_dets.element.xpath(ASSIGN_LISTCOMP_XPATH)
    brief_msg = ''
    plural = 's' if len(listcomp_els) > 1 else ''
    for i, dict_el in enumerate(listcomp_els):
        first = (i == 0)
        name = get_assign_name(dict_el)
        items = code_execution.get_val(
            block_dets.pre_block_code_str, block_dets.block_code_str, name)
        if first:
            title = layout(f"""\

                ### List comprehension{plural} used

                """)
            brief_msg += title
        brief_msg += layout(f"""

            `{name}` is a list comprehension returning a list
            with {utils.int2nice(len(items))} items: {items}
            """)
    if repeated_message:
        extra_msg = ''
    else:
        extra_msg = (
            layout(f"""\
                ### Other "comprehensions"

                """)
            + shared.GENERAL_COMPREHENSION_COMMENT
            + layout("""\

                List comprehensions aren't the only type of comprehension you can
                make. Python also lets you write Dictionary and Set Comprehensions:

                """)
            + shared.DICT_COMPREHENSION_COMMENT
            + '\n\n'
            + shared.SET_COMPREHENSION_COMMENT
            + '\n\n'
            + layout("""\
                Pro tip: don't make comprehension *in*comprehensions ;-). If it is
                hard to read it is probably better written as a looping structure.
                """)
        )
    message = {
        conf.BRIEF: brief_msg,
        conf.EXTRA: extra_msg,
    }
    return message



from . import cli_colour, cli_conf, cli_utils

def _get_vertical_padding_line(length):
    return length * ' '

def h(text, level):
    level_colour = cli_conf.LEVEL2COLOUR.get(level)
    bold = False
    if level == 1:
        text = text.center(cli_conf.TERMINAL_WIDTH)
        vertical_padding_line = _get_vertical_padding_line(
            length=cli_conf.TERMINAL_WIDTH)
    else:
        if level == 3:
            lstar_padding = '**** '
            rstar_padding = ' ****'
        else:
            lstar_padding, rstar_padding = '', ''
        text = f"{lstar_padding}{text}{rstar_padding} "  ## extra space at end so right side of reversed background has a space between it and the text
        vertical_padding_line = _get_vertical_padding_line(length=len(text))
    if level <= 2:
        text = f"{vertical_padding_line}\n{text}\n{vertical_padding_line}"
        bold = True
    return '\n' + cli_colour.colourise(
        text, level_colour, reverse=True, bold=bold)

def h1(s, **_kwargs):
    return h(s, level=1)

def h2(s, **_kwargs):
    return h(s, level=2)

def h3(s, **_kwargs):
    return h(s, level=3)

def h4(s, **_kwargs):
    return h(s, level=4)

def h5(s, **_kwargs):
    return h(s, level=5)

def p(text, **_kwargs):
    return cli_colour.colourise(text, cli_conf.T)

def a(text, **_kwargs):
    return cli_colour.colourise_low_vis(text)

def hr(_text, **kw):
    """
    Horizontal line (rule) that has colour of appropriate heading level.
    """
    nesting_level = kw.get('nesting_level', 1)
    indent = (nesting_level - 1) * cli_conf.LEFT_INDENT
    colour = cli_conf.LEVEL2COLOUR[nesting_level]
    start = end = cli_colour.colourise(cli_conf.HR_ENDS, colour)
    return cli_colour.colourise_low_vis(
        f"\n{indent}{start}{cli_conf.HR_MARKER}{end}{indent}\n")

def code(text, from_fenced_block=None, **kw):
    """
    md code AND ``` style fenced raw code ends here
    """
    if not from_fenced_block:
        text = ('\n' + text).replace('\n    ', '\n')[1:]
    # funny: ":-" confuses the tokenizer. replace/backreplace:
    raw_code = text.replace(':-', '\x01--')
    text = cli_utils.style_ansi(raw_code)
    # unnested level has indent of N_LEFT_INDENT, use it for fenced
    indent = ' ' * kw.get('nesting_level', cli_conf.N_LEFT_INDENT)
    # if from_fenced_block: ... WE treat equal.
    # shift to the far left, no matter the indent (screen space matters):
    firstl = text.split('\n')[0]
    n_spaces2delete = len(firstl) - len(firstl.lstrip())
    spaces2delete = ' ' * n_spaces2delete
    text = '\n' + (f"\n{text}").replace(f"\n{spaces2delete}", '\n')[1:]
    # we want an indent of one and low vis prefix. this does it:
    code_lines = text.splitlines()
    code_prefix = cli_colour.colourise_low_vis(cli_conf.CODE_PREFIX)
    empty = cli_colour.colourise('', cli_conf.CODE_COLOUR, no_reset=True)
    prefix = f"\n{indent}{code_prefix} {empty}"
    if code_lines[-1] == '\x1b[0m':
        code_lines.pop()
    code = prefix.join(code_lines)
    code = code.replace('\x01--', ':-')
    return code + '\n' + cli_conf.DEFAULT_ANSI_COLOUR_BYTE_STR



from collections import defaultdict, namedtuple, Counter

from ..advisors import filt_block_advisor
from .. import conf
from ..utils import get_python_version, int2nice, layout_comment as layout

def get_string_3_7(comparison_el):
    string = comparison_el.get('s')
    if not string:
        return None
    return string

def get_string_3_8(comparison_el):
    val = comparison_el.get('value')
    if comparison_el.get('type') == 'str':
        string = val
    else:
        string = None
    return string

def get_num_3_7(comparison_el):
    num = comparison_el.get('n')
    if not num:
        return None
    return num

def get_num_3_8(comparison_el):
    val = comparison_el.get('value')
    if comparison_el.get('type') in ('int', 'float'):
        num = val
    else:
        num = None
    return num

python_version = get_python_version()
if python_version in (conf.PY3_6, conf.PY3_7):
    get_string = get_string_3_7
    get_num = get_num_3_7
elif python_version == conf.PY3_8:
    get_string = get_string_3_8
    get_num = get_num_3_8
else:
    raise Exception(f"Unexpected Python version {python_version}")

IfDets = namedtuple('IfDetails',
    'multiple_conditions, missing_else, if_clauses')

ELIF = 'elif'
ELSE = 'else'

STRING = 'string'
NUMBER = 'number'

IF_XPATH = 'descendant-or-self::If'

def add_if_details(if_element, if_clauses):
    """
    If under our <If>'s <orelse> we have another <If>, and it is the only child,
    store an 'elif' in if_details and send the <If> under <orelse> through
    again (the AST nests them even though they are siblings in the original code
    syntax); otherwise store an 'else' in if_details and return.

    <If>s have <orelse>s which either have sole <If>s or not. When there is not
    then we have reached else.
    """
    orelse_el = if_element.xpath('orelse')[0]  ## always has one
    orelse_children = orelse_el.getchildren()
    if not orelse_children:
        return  ## merely an If on its own without clauses
    ## else: if or elif: ?
    just_the_if_under_orelse = (
        len(orelse_children) == 1 and orelse_children[0].tag == 'If')
    has_elif = just_the_if_under_orelse
    if has_elif:
        if_clauses.append(ELIF)
        elif_el = orelse_children[0]
        add_if_details(elif_el, if_clauses)
    else:
        if_clauses.append(ELSE)
    return

def get_ifs_details(block_dets):
    """
    There can be multiple if statements in a snippet so we have to handle each
    of them.

    And in AST the following are very similar:

    if foo:
        ...
    else:
        if bar:
            ...

    if foo:
        ...
    elif bar:
        ...

    They both result in:
    <orelse>
        <If ...>

    So need to look under the 'orelse' for anything other than an 'If' to decide
    whether an 'else' or 'elif'.
    """
    ## skip when if __name__ == '__main__'
    if block_dets.block_code_str.startswith("if __name__ == "):
        return []
    raw_if_els = block_dets.element.xpath(IF_XPATH)
    if_elements = []
    for raw_if_el in raw_if_els:
        ## ignore if really an elif
        parent_el = raw_if_el.getparent()
        has_or_else_parent = (parent_el.tag == 'orelse')
        has_siblings = bool(parent_el.getchildren())
        actually_elif = (has_or_else_parent and not has_siblings)
        if not actually_elif:
            if_elements.append(raw_if_el)
    ifs_details = []
    for if_element in if_elements:
        if_clauses = []
        add_if_details(if_element, if_clauses)
        multiple_conditions = bool(if_clauses)
        if multiple_conditions:
            last_sub_if = if_clauses[-1]
            missing_else = (last_sub_if != ELSE)
        else:
            missing_else = False
        ifs_details.append(
            IfDets(multiple_conditions, missing_else, if_clauses))
    return ifs_details

def _get_if_comment(ifs_details):
    """
    Have to cope with multiple if statements and make it nice (unnumbered) when
    only one.
    """
    brief_msg = '### Conditional statement detected'
    for n, if_details in enumerate(ifs_details, 1):
        counter = '' if len(ifs_details) == 1 else f" {int2nice(n)}"
        if if_details.multiple_conditions:
            n_elifs = if_details.if_clauses.count(ELIF)
            brief_msg += layout(f"""\

                `if` statement{counter} has {int2nice(n_elifs)} `elif` clauses
                """)
            if if_details.missing_else:
                brief_msg += " and no `else` clause."
            else:
                brief_msg += " and an `else` clause."
            brief_msg += (
                " Note - `else` clauses with an `if`, and only an `if`, "
                "underneath count as `elif` clauses not `else` clauses.")
        else:
            brief_msg += layout(f"""\

                `if` statement{counter} has no extra clauses e.g. `elif`s or an
                `else`.
                """)
    return brief_msg

@filt_block_advisor(xpath=IF_XPATH)
def if_else_overview(block_dets, *, repeated_message=False):
    """
    Look at conditional statements using if (apart from if __name__ ==
    '__main__").
    """
    ## skip when if __name__ == '__main__'
    if block_dets.block_code_str.startswith("if __name__ == "):
        return None
    ifs_details = get_ifs_details(block_dets)
    brief_msg = _get_if_comment(ifs_details)
    main_msg = brief_msg
    if not repeated_message:
        main_msg += (
            layout("""\

                When using `if`, `elif`, or `else`, the item evaluated can
                either be an expression or a name (variable). Sometimes your
                code can be more readable if the expression is precalculated and
                a name is supplied instead. For example:

                """)
            +
            layout("""\
                if length_matters and len(phrase) > conf.MAX_SANE_LENGTH:
                    phrase = chop(phrase)
                """, is_code=True)
            +
            layout("""\

                is shorter but more complex than:

                """)
            +
            layout("""\
                too_long = (
                    length_matters
                    and len(phrase) > conf.MAX_SANE_LENGTH
                )
                if too_long:
                    phrase = chop(phrase)
                """, is_code=True)
        )
    message = {
        conf.BRIEF: brief_msg,
        conf.MAIN: main_msg,
    }
    return message

@filt_block_advisor(xpath=IF_XPATH, warning=True)
def missing_else(block_dets, *, repeated_message=False):
    """
    Warn about benefits in many cases of adding else clause if missing.
    """
    ifs_details = get_ifs_details(block_dets)
    brief_msg = ''
    has_missing_else = False
    ifs_dets_missing_else = [
        if_details for if_details in ifs_details if if_details.missing_else]
    for n, if_details in enumerate(ifs_dets_missing_else, 1):
        first = (n == 1)
        counter = '' if len(ifs_dets_missing_else) == 1 else f" {int2nice(n)}"
        if first:
            brief_msg += layout(f"""\

                ### Possibly better with `else` clause

                """)
        if first and not repeated_message:
            brief_msg += layout(f"""\
                `if` block{counter} has `elif` clauses but lacks an `else`
                clause. If your `elif` clauses are trying to handle all expected
                cases it is probably best to include an `else` clause as well
                just in case something unexpected happens.

                Note - `else` clauses with an `if` and only the `if` underneath
                count as `elif` clauses.
                """)
        has_missing_else = True
    if not has_missing_else:
        return None
    if repeated_message:
        main_msg = brief_msg
    else:
        main_msg = (
            brief_msg
            +
            layout("""\

                You may have left out the `else` because it is currently
                impossible that this branch will ever be called. You know that
                you can only receive the items currently handled by the `elif`s
                so `else` can logically never be called. And that it true -
                until it isn't later ;-) - for example, if the calling code
                starts supplying more types of whatever is being evaluated in
                the `elif`s. This problem happens surprisingly often and can
                create nasty bugs that are hard to trace. If you add an `else`
                clause that raises an exception you will instantly know if the
                expected conditions for your conditional are breached and
                exactly what to fix.

                For example:

                """)
            +
            layout("""\
                ## At the point this code is written we absolutely know
                ## there are only two user types: managers and staff

                if user == 'manager':
                    create_alert()
                    set_up_manager(user)
                elif user == 'staff':
                    set_up_staff(user)
                else:
                    ## This can logically never happen so
                    ## we might be tempted to leave out the else clause
                    pass
                """, is_code=True)
            +
            layout("""\

                But then a new user type is created (e.g. admin) and the logic
                of the entire program breaks all over the place with no clarity
                about the real source.

                It would be better to have an `else` clause as follows so you
                know exactly what violated the program's assumptions:

                """)
            +
            layout("""\
                else:
                    raise Exception(f"Unexpected user: '{user}'")
                """, is_code=True)
        )
    message = {
        conf.BRIEF: brief_msg,
        conf.MAIN: main_msg,
    }
    return message

def get_split_membership_dets(if_el):
    """
    if x == 'a' or x == 'b' or x == 'c':
        print(x)

    If/test/BoolOp/values/Compare
                                 left/Name id 'x'
                                 comparators/Constant s 'a'
                                 comparators/Num n  etc

    Only provide message using content if all items are of the same type and are
    either numbers or strings.
    """
    compare_els = if_el.xpath('test/BoolOp/values/Compare')
    if not compare_els:
        return None
    left_name_comp_vals = defaultdict(list)
    basic_types = set()
    for compare_el in compare_els:
        left_name_els = compare_el.xpath('left/Name')
        if not left_name_els:
            continue
        left_name = left_name_els[0].get('id')
        if not left_name:
            continue
        comparators_els = compare_el.xpath('comparators')
        if not comparators_els:
            continue
        comparators_el = comparators_els[0]
        comparison_els = comparators_el.getchildren()  ## e.g. Strs, Nums, NameConstants etc
        if not comparison_els:
            continue
        comparison_el = comparison_els[0]  ## Str or Num etc
        comp_val = get_string(comparison_el)
        if comp_val is not None:
            basic_types.add(STRING)
            if len(basic_types) > 1:
                return None
        else:
            comp_val = get_num(comparison_el)
            if comp_val is not None:
                basic_types.add(NUMBER)
                if len(basic_types) > 1:
                    return None
            else:
                return None  ## Give up - something is not right
        left_name_comp_vals[left_name].append(comp_val)
    c = Counter(left_name_comp_vals.keys())
    comp_var = c.most_common(1)[0][0]
    comp_vals = left_name_comp_vals[comp_var]
    comp_val_strs = []
    basic_type = basic_types.pop()
    for comp_val in comp_vals:
        if basic_type == STRING:
            quoter = "'" if '"' in comp_val else "'"
            comp_val_strs.append(f"{quoter}{comp_val}{quoter}")
        elif basic_type == NUMBER:
            comp_val_strs.append(str(comp_val))
        else:
            raise ValueError(f"Unexpected basic type: '{basic_type}'")
    comp_vals_gp_str = '[' + ', '.join(comp_val_strs) + ']'
    return comp_var, comp_vals_gp_str

@filt_block_advisor(xpath=IF_XPATH)
def split_group_membership(block_dets, *, repeated_message=False):
    """
    Explain how to use in group and not in group rather than multiple
    comparisons.

    if x == 'a' or x == 'b' or x == 'c':
        print(x)
    ==>
    if x in ['a', 'b', 'c']:
        print(x)
    """
    if_els = block_dets.element.xpath(IF_XPATH)
    has_split = False
    for if_el in if_els:
        try:
            comp_var, comp_vals_gp_str = get_split_membership_dets(if_el)
        except TypeError:
            continue
        else:
            has_split = True
            break
    if not has_split:
        return None
    brief_msg = layout(f"""\

            ### Possible option of evaluating group membership

            It looks like `{comp_var}` has been in multiple separate
            comparisons. In Python it is possible to do a simple check of group
            membership instead.
        """)
    if not repeated_message:
        brief_msg += (
            layout(f"""\

                For example:

            """)
            +
            layout(f"""\
                if {comp_var} in {comp_vals_gp_str}:
                    ...
                """, is_code=True)
        )
    main_msg = brief_msg
    if not repeated_message:
        main_msg += (
            layout(f"""\

                Or to check if a variable is NOT in a group:

                """)
            +
            layout(f"""\
                if {comp_var} not in {comp_vals_gp_str}:
                    ...
                """, is_code=True)
        )
    message = {
        conf.BRIEF: brief_msg,
        conf.MAIN: main_msg,
    }
    return message

def get_has_explicit_count(if_el):
    compare_els = if_el.xpath('test/Compare')
    if not compare_els:
        return False
    compare_el = compare_els[0]
    func_name_els = compare_el.xpath('left/Call/func/Name')
    if not func_name_els:
        return False
    len_func = (func_name_els[0].get('id') == 'len')
    if not len_func:
        return False
    ops_els = compare_el.xpath('ops')
    if not ops_els:
        return False
    ops_el = ops_els[0]
    operator_els = ops_el.getchildren()
    if not operator_els:
        return False
    operator_type = operator_els[0].tag  ## e.g. Gt
    comparators_els = compare_el.xpath('comparators')
    if not comparators_els:
        return False
    comparator_el = comparators_els[0]
    comparison_els = comparator_el.getchildren()
    if not comparison_els:
        return False
    comparison_el = comparison_els[0]
    n = get_num(comparison_el)
    if n is None:
        return False
    explicit_booleans = [
        ('Gt', '0'),
        ('GtE', '1'),
        ('Eq', '0'),
        ('LtE', '0'),
        ('Lt', '1'),
    ]
    has_explicit_boolean = ((operator_type, n) in explicit_booleans)
    return has_explicit_boolean

@filt_block_advisor(xpath=IF_XPATH)
def implicit_boolean_enough(block_dets, *, repeated_message=False):
    """
    Look for cases where an implicit boolean comparison is enough.
    """
    if repeated_message:
        return None
    if_els = block_dets.element.xpath(IF_XPATH)
    implicit_boolean_possible = False
    for if_el in if_els:
        has_explicit_count = get_has_explicit_count(if_el)
        if has_explicit_count:
            implicit_boolean_possible = True
        else:
            continue
    if not implicit_boolean_possible:
        return None
    brief_msg = (
        layout("""\

            ### Possible option of using an implicit boolean

            In Python, "", 0, [], {}, (), `None` all evaluate to False when we
            ask if they are `True` or `False`. And they all evaluate to `True`
            if they contain something. WAT?!

            Well, if we have a list `my_list` we can replace:

            """)
        +
        layout("""\
            if len(my_list) > 0:
                ...

            """, is_code=True)
        +
        layout("""\
            with:

            """)
        +
        layout("""\
            if my_list:  ## if empty, evaluates to False otherwise True
                ...

            """, is_code=True)
    )
    message = {
        conf.BRIEF: brief_msg,
    }
    return message



from ..advisors import filt_block_advisor
from ..ast_funcs import get_assign_name
from .. import code_execution, conf, utils
from ..utils import layout_comment as layout

ASSIGN_LIST_XPATH = 'descendant-or-self::Assign/value/List'

DEFAULT_EXAMPLE_TYPE = conf.STR_TYPE

def get_item_type_names(items):
    item_type_names = sorted(set(
        [type(item).__name__ for item in items]
    ))
    item_type_nice_names = [
        conf.TYPE2NAME.get(item_type, item_type)
        for item_type in item_type_names]
    return item_type_names, item_type_nice_names

def get_type_for_example(list_items):
    """
    If all else fails, just use strings for examples.
    """
    if list_items is None:
        type4example = DEFAULT_EXAMPLE_TYPE
    else:
        item_type_names, _item_type_nice_names = get_item_type_names(list_items)
        try:
            type4example = item_type_names.pop()
        except IndexError:
            type4example = DEFAULT_EXAMPLE_TYPE
    return type4example

def _get_additional_main_msg(first_name, first_list_items):
    """
    :param list first_list_items: note - may be None if a list is defined in a
     function
    """
    type4example = get_type_for_example(first_list_items)
    try:
        example_items = conf.EXAMPLES_OF_TYPES[type4example]
    except KeyError:
        example_items = conf.EXAMPLES_OF_TYPES[DEFAULT_EXAMPLE_TYPE]
    example_item = example_items[0]
    listable_example_item = (f"'{example_item}'"
        if type4example == conf.STR_TYPE else example_item)
    if first_list_items is None:
        appended_list = []
    else:
        appended_list = first_list_items.copy()
    appended_list.append(example_item)
    extended_list = appended_list.copy()
    items2extend = example_items[1:3]  ## don't repeat the appended item - might confuse the user at this stage
    extended_list.extend(items2extend)
    friends = ['Selma', 'Willy', 'Principal Skinner']
    family = ['Bart', 'Lisa', 'Marge', 'Homer']
    guests = friends + family
    additional_main_msg = (
        layout("""\

            Lists, along with dictionaries, are the workhorses of Python data
            structures.

            Lists have an order, and can contain duplicate items and items of
            different types (usually not advisable).

            Extra items can be added to lists using the .append() method e.g.
            """)
        +
        layout(f"""\
            {first_name}.append({listable_example_item})
            """, is_code=True)
        +
        layout(f"""\

            which results in {appended_list}

            """)
        +
        layout("""\

            If you want to add multiple items at once, .extend() is useful.

            Note - we are extending the list which has already had an item
            appended, not the original list.

            """)
        +
        layout(f"""\
            {first_name}.extend({items2extend})
            """, is_code=True)
        +
        layout(f"""\

            which results in {extended_list}

            """)
        +
        layout("""\

            GOTCHA: if you are adding tuples to your list it is easy to
            forget the nested parentheses. E.g.

            """)
        +
        layout(f"""\

            coordinates.append((x, y))  ## Correct

            coordinates.append(x, y)  ## Oops - append only takes one item not two

            """, is_code=True)
        +
        layout("""\

            Lists can also be added together e.g.

            """)
        +
        layout(f"""\
            friends = {friends}
            family = {family}
            guests = friends + family

            """, is_code=True)
        +
        layout(f"""\

            resulting in {guests}

            GOTCHA: you can't add individual items to lists unless you put
            them in a list as well E.g.

            """)
        +
        layout(f"""\
            workmate = 'Carl'
            guests = {friends} + {family} + workmate  ## Oops - can only add lists
            guests = {friends} + {family} + [workmate]  ## That's better

            """, is_code=True)
        +
        layout(f"""\

            resulting in {guests + ['Carl']}

            """)
        )
    return additional_main_msg

## only interested in lists when being assigned as a value
## (i.e. <body><Assign><value><List> so we're looking for List under value only)
@filt_block_advisor(xpath=ASSIGN_LIST_XPATH)
def list_overview(block_dets, *, repeated_message=False):
    """
    General overview of list taking content details into account.
    """
    list_els = block_dets.element.xpath(ASSIGN_LIST_XPATH)
    brief_msg = ''
    main_msg = ''
    plural = 's' if len(list_els) > 1 else ''
    first_name = None
    first_list_items = None
    for i, list_el in enumerate(list_els):
        first = (i == 0)
        name = get_assign_name(list_el)
        items = code_execution.get_val(  ## might be None if defined in a function
            block_dets.pre_block_code_str, block_dets.block_code_str, name)
        if first:
            first_name = name
            first_list_items = items
            title = layout(f"""\

                ### List{plural} defined
    
                """)
            brief_msg += title
            main_msg += title
        if items is None:
            list_desc = layout(f"""\

                `{name}` is a list.
                """)
        elif items == []:
            list_desc = layout(f"""\

                `{name}` is an empty list.
                """)
        else:
            list_desc = layout(f"""\
    
                `{name}` is a list with {utils.int2nice(len(items))} items.
                """)
        brief_msg += list_desc
        main_msg += list_desc
    if not repeated_message:
        brief_msg += layout("""\

            Lists, along with dictionaries, are the workhorses of Python data
            structures.

            Lists have an order, and can contain duplicate items and items of
            different types (usually not advisable).
            """)
        main_msg += _get_additional_main_msg(
            first_name, first_list_items)
    message = {
        conf.BRIEF: brief_msg,
        conf.MAIN: main_msg,
    }
    return message

@filt_block_advisor(xpath=ASSIGN_LIST_XPATH, warning=True)
def mixed_list_types(block_dets, *, repeated_message=False):  # @UnusedVariable
    """
    Warns about lists containing a mix of data types.
    """
    list_els = block_dets.element.xpath(ASSIGN_LIST_XPATH)
    brief_msg = ''
    main_msg = ''
    has_mixed = False
    for i, list_el in enumerate(list_els):
        first = (i == 0)
        name = get_assign_name(list_el)
        items = code_execution.get_val(  ## might be None if defined in a function
            block_dets.pre_block_code_str, block_dets.block_code_str, name)
        if not items:
            continue
        _item_type_names, item_type_nice_names = get_item_type_names(items)
        if len(item_type_nice_names) <= 1:
            ## No explanation needed if there aren't multiple types.
            continue
        has_mixed = True
        if first:
            title = layout(f"""\
                ### List(s) with mix of different data types
    
                """)
            brief_msg += title
            main_msg += title
        mixed_warning = layout(f"""

            `{name}` contains more than one data type - which is probably a bad
            idea.
            """)
        brief_msg += mixed_warning
        main_msg += mixed_warning
        main_msg += layout(f"""\
             The data types found were: {", ".join(item_type_nice_names)}.
            """)
    if not has_mixed:
        return None
    message = {
        conf.BRIEF: brief_msg,
        conf.MAIN: main_msg,
    }
    return message



from ..advisors import filt_block_advisor
from .. import ast_funcs, conf
from ..utils import layout_comment as layout

FOR_XPATH = 'descendant-or-self::For'
WHILE_XPATH = 'descendant-or-self::While'
IF_XPATH = 'descendant-or-self::If'  ## includes elif
OUTER_XPATHS = {
    'for': FOR_XPATH,
    'while': WHILE_XPATH,
    'if': IF_XPATH,
}
NESTING_XPATH = ' | '.join(OUTER_XPATHS.values())

def _long_nested_block(outer_el):
    _first_line_no, _last_line_no, el_lines_n = ast_funcs.get_el_lines_dets(
        outer_el, ignore_trailing_lines=True)
    nested_block_line_len = el_lines_n - 1  ## ignore the actual outer el e.g. for:
    too_long = nested_block_line_len > conf.MAX_BRIEF_NESTED_BLOCK
    return too_long

def has_long_block(block_el, xpath):
    long_block = False
    outer_els = block_el.xpath(xpath)
    for outer_el in outer_els:
        too_long = _long_nested_block(outer_el)
        if too_long:
            long_block = True
            break
    return long_block

@filt_block_advisor(xpath=NESTING_XPATH, warning=True)
def bloated_nested_block(block_dets, *, repeated_message=False):
    """
    Look for long indented blocks under conditionals, inside loops etc that are
    candidates for separating into functions to simplify the narrative of the
    main code.
    """
    bloated_outer_types = set()
    for lbl, outer_xpath in OUTER_XPATHS.items():
        if has_long_block(block_dets.element, outer_xpath):
            bloated_outer_types.add(lbl)
    if not bloated_outer_types:
        return None
    brief_msg = layout("""\
        
        ### Possible option of replacing long nested block with function call

        """)
    for bloated_outer_type in bloated_outer_types:
        brief_msg += layout(f"""\
            The code has at least one long nested block under
            `{bloated_outer_type}:`
            """)
    if repeated_message:
        main_msg = brief_msg
        extra_msg = ''
    else:  ## Hypocrisy alert LOL - I complain here about excessive nested blocks
        brief_msg += layout("""\

            It may be possible to pull most of the nested code block into a
            function which can be called instead.
            """)
        main_msg = brief_msg
        main_msg += (
            layout("""

                For example, instead of:

                """)
            +
            layout("""\
                for name in names:
                    ## contact name
                    line 1
                    line 2
                    line 3
                    ...
                    line 30
                logging.info("Finished!")
                """, is_code=True)
            +
            layout("""\

                you could possibly write:

                """)
            +
            layout('''\
                def contact(name):
                    """
                    Contact person ...
                    """
                    line 1
                    line 2
                    line 3
                    ...
                    line 30

                for name in names:
                    contact(name)
                logging.info("Finished!")
                ''', is_code=True)
        )
        extra_msg = layout("""\
            Computers can handle lots of nesting without malfunctioning. Human
            brains are not so fortunate. As it says in The Zen of Python:

            > "Flat is better than nested."
            """)
    message = {
        conf.BRIEF: brief_msg,
        conf.MAIN: main_msg,
        conf.EXTRA: extra_msg,
    }
    return message



from ..advisors import any_block_advisor, filt_block_advisor
from .. import code_execution, conf
from ..utils import get_python_version, layout_comment as layout

F_STR = 'f-string'
STR_FORMAT_FUNC = 'str_format'
SPRINTF = 'sprintf'
STR_ADDITION = 'string addition'

ASSIGN_VALUE_XPATH = 'descendant-or-self::Assign/value'
FUNC_ATTR_XPATH = 'descendant-or-self::value/Call/func/Attribute'
JOINED_STR_XPATH = 'descendant-or-self::Assign/value/JoinedStr'
SPRINTF_XPATH = 'descendant-or-self::value/BinOp/op/Mod'
STR_ADDITION_XPATH = 'descendant-or-self::BinOp/left/Str'  ## each left has a right as well so only need to look at one to get at binop and assign ancestors etc

F_STR_REMINDER = False

def get_str_els_3_7(block_el):
    str_els = block_el.xpath('descendant-or-self::Assign/value/Str')
    return str_els

def get_str_els_3_8(block_el):
    assign_val_els = block_el.xpath(ASSIGN_VALUE_XPATH)
    str_els = []
    for assign_val_el in assign_val_els:
        assign_str_els = assign_val_el.xpath('Constant')
        if len(assign_str_els) != 1:
            continue
        assign_str_el = assign_str_els[0]
        if assign_str_el.get('type') == 'str':
            str_els.append(assign_str_el)
    return str_els

def get_str_els_being_combined_3_7(block_el):
    str_els_being_combined = block_el.xpath(
        'descendant-or-self::BinOp/left/Str')
    return str_els_being_combined

def get_str_els_being_combined_3_8(block_el):
    left_str_els = block_el.xpath('descendant-or-self::BinOp/left/Constant')
    str_els_being_combined = []
    for left_str_el in left_str_els:
        if left_str_el.get('type') == 'str':
            str_els_being_combined.append(left_str_el)
    return str_els_being_combined

python_version = get_python_version()
if python_version in (conf.PY3_6, conf.PY3_7):
    get_str_els = get_str_els_3_7
    get_str_els_being_combined = get_str_els_being_combined_3_7
elif python_version == conf.PY3_8:
    get_str_els = get_str_els_3_8
    get_str_els_being_combined = get_str_els_being_combined_3_8
else:
    raise Exception(f"Unexpected Python version {python_version}")

@filt_block_advisor(xpath=ASSIGN_VALUE_XPATH)
def assigned_str_overview(block_dets, *, repeated_message=False):
    """
    Provide overview of assigned strings e.g. name = 'Hamish'.
    """
    brief_msg = ''
    str_els = get_str_els(block_dets.element)
    if not str_els:
        return None
    first_name = None
    first_val = None
    for i, str_el in enumerate(str_els):
        assign_el = str_el.xpath('ancestor::Assign')[0]
        first = (i == 0)
        name = assign_el.xpath('targets/Name')[0].get('id')
        val = code_execution.get_val(
            block_dets.pre_block_code_str, block_dets.block_code_str, name)
        if first:
            first_name = name
            first_val = val
            brief_msg += layout("""\

                #### String Overview

                """)
        brief_msg += layout(f"""\
            `{name}` is a string.

            """)
    if repeated_message:
        main_msg = brief_msg
        extra_msg = ''
    else:
        brief_msg += layout(f"""\
            Python makes it easy to do lots of cool things with strings.

            """)
        if first_val:
            name2use = first_name
            val2use = first_val
        else:
            name2use = 'address'
            val2use = 'Waiuku, New Zealand'
            brief_msg += layout(f"""\
                For illustration, imagine we have string '{val2use}' assigned to
                `{name2use}`:

                """)
        main_msg = brief_msg  ## after this brief and main diverge
        brief_msg += layout(f"""\
            {name2use}.upper() returns {val2use.upper()}.

            """)
        black_heart = "\N{BLACK HEART}"
        main_msg += layout(f"""\
            Examples:

            {name2use}.upper() returns {val2use.upper()}.

            {name2use}.center(70, '=') returns {val2use.center(70, '=')}

            {name2use}.endswith('chicken') returns {val2use.endswith('chicken')}

            {name2use} + ' is a string' returns {val2use + ' is a string'}

            {name2use} + ' ' + '{{\\NBLACK HEART}}' + ' Python' returns
            {val2use + ' ' + black_heart + ' Python'}

            len({name2use}) returns {len(val2use)} because that is how many
            characters are in the {name2use} string (remember to count spaces -
            they are characters too)

            sorted({name2use}) returns {sorted(val2use)}
            """)
        extra_msg = layout("""\
            .upper(), .center() etc are abilities available with all Python
            strings. Technically they are methods of string objects. They start
            with a dot and are on the end of the object.

            To see the full list of string methods enter dir(str) into a Python
            command line.

            len() is a function which can be used on lots of things - not just
            string objects. It is not a method of the string object. Other
            functions that are not string-specific but are commonly used with
            strings include sorted() and print().
            """)
    message = {
        conf.BRIEF: brief_msg,
        conf.MAIN: main_msg,
        conf.EXTRA: extra_msg,
    }
    return message

def str_combination(combination_type, str_els, *, repeated_message=False):
    global F_STR_REMINDER
    brief_msg = ''
    title = None
    for str_el in str_els:
        try:
            assign_el = str_el.xpath('ancestor::Assign')[-1]
        except IndexError:
            name = "An unnamed string"
        else:
            raw_name = assign_el.xpath('targets/Name')[0].get('id')
            name = f"`{raw_name}`"
        combination_type2comment = {
            F_STR: "f-string interpolation",
            STR_FORMAT_FUNC: "the format function",
            SPRINTF: "sprintf string interpolation",
            STR_ADDITION: "string addition (e.g. animal = 'jelly' + 'fish')",
        }
        combination_comment = combination_type2comment[combination_type]
        if not title:
            title = layout(f"""\

            ### Strings created by combining or interpolating strings
            """)
            brief_msg += title
        brief_msg += layout(f"""\

            {name} is created using {combination_comment}.
            """)
    message = {
        conf.BRIEF: brief_msg,
        conf.MAIN: brief_msg,
    }
    if combination_type != F_STR and not repeated_message:
        if not F_STR_REMINDER:
            F_STR_REMINDER = True
            message[conf.BRIEF] += layout("""\
    
                Your snippet uses a non-f-string approach to constructing a
                string.

                Have you considered using an f-string approach to constructing
                your string?
                """)
            message[conf.MAIN] += (
                layout("""\

                    Have you considered using an f-string approach to
                    constructing your string?

                    f-strings let you reference variables from earlier in your
                    code and allow very readable string construction. All the
                    usual tricks of the .format() approach also work. For
                    example, comma separating thousands in numbers:

                    """)
                +
                layout("""\
                    cost = 10_550
                    print(f"Cost is ${cost:,}")
                    # >>> 'Cost is $10,550'
                    """, is_code=True)
                +
                layout(f"""\

                    Or making small in-line calculations:

                    """)
                +
                layout("""\
                    people = ['Bart', 'Lisa']
                    print(f"There are {len(people)} people")
                    # >>> 'There are 2 people'
                    """, is_code=True)
                +
                layout(f"""\

                    Or zero-padding numbers:

                    """)
                +
                layout("""\
                    num = 525
                    print(f"{num:0>4}")
                    # >>> '0525'
                    """, is_code=True)
            )
    return message

@filt_block_advisor(xpath=JOINED_STR_XPATH)
def f_str_interpolation(block_dets, *, repeated_message=False):
    """
    Examine f-string interpolation.
    """
    joined_els = block_dets.element.xpath(JOINED_STR_XPATH)
    return str_combination(F_STR,
        joined_els, repeated_message=repeated_message)

@filt_block_advisor(xpath=FUNC_ATTR_XPATH)
def format_str_interpolation(block_dets, repeated_message=False):
    """
    Look at use of .format() to interpolate into strings.
    """
    func_attr_els = block_dets.element.xpath(FUNC_ATTR_XPATH)
    format_funcs = []
    for func_attr_el in func_attr_els:
        is_format_func = func_attr_el.get('attr') == 'format'
        if is_format_func:
            format_funcs.append(func_attr_el)
    if not format_funcs:
        return None
    return str_combination(STR_FORMAT_FUNC,
        format_funcs, repeated_message=repeated_message)

@any_block_advisor()
def sprintf(block_dets, *, repeated_message=False):
    """
    Look at use of sprintf for string interpolation e.g. greeting = "Hi %s" %
    name
    """
    sprintf_els = block_dets.element.xpath(SPRINTF_XPATH)
    has_sprintf = bool(sprintf_els)
    if not has_sprintf:
        return None
    return str_combination(SPRINTF,
        sprintf_els, repeated_message=repeated_message)

@any_block_advisor()
def string_addition(block_dets, *, repeated_message=False):
    """
    Advise on string combination using +. Explain how f-string alternative
    works.
    """
    str_els_being_combined = get_str_els_being_combined(block_dets.element)
    has_string_addition = False
    str_addition_els = []
    for str_el in str_els_being_combined:
        ## Does their immediate ancestor BinOp have op of Add?
        ## Don't know if there any alternatives but let's be sure
        ## Ordered set of nodes, from parent to ancestor? https://stackoverflow.com/a/15645846
        bin_op_el = str_el.xpath('ancestor-or-self::BinOp')[-1]
        ## was it an Add op
        has_add = bool(bin_op_el.xpath('op/Add'))
        if has_add:
            str_addition_els.append(str_el)
            has_string_addition = True
    if not has_string_addition:
        return None
    addition_message = str_combination(STR_ADDITION,
        str_addition_els, repeated_message=repeated_message)
    return addition_message



import logging

def get_val(pre_block_code_str, block_code_str, name):
    """
    Executing supplied code from end users - nope - nothing to see here from a
    security point of view ;-) Needs addressing if this code is ever used as a
    service for other users.

    Note - can be the source of mysterious output in stdout (e.g. exec a print
    function).
    """
    exp_dets = {}
    try:
        exec(pre_block_code_str + block_code_str, exp_dets)
    except ImportError as e:
        logging.debug(
            f"Import problem running {__file__} (specifically {__name__}): {e}")
        raise ImportError("SuperHELP only has modules from the Python standard "
            "library installed - it looks like your snippet relies on a module "
            "from outside the standard library.")
    try:
        val = exp_dets[name]
    except KeyError:
        val = None
#         raise KeyError(
#             f"Unable to find name '{name}' in code_str\n{block_code_str}")
    return val




from . import cli_conf

def colourise(text, colour, *, reverse=False, bold=False, no_reset=False):
    """
    Colour is toggled on and will keep being applied unless toggled off (back to
    default colour).

    We apply this colour to everything except for nested content. That gets its
    own colours depending on what type of thing it is e.g. code.

    The only reason you'd supply an empty string to this is if you were setting
    a colour and had no_reset=True. Otherwise an empty line only would be
    coloured which makes no sense.

    Testing: print('Should be red spam', colourise('spam', cli_conf.RED))

    :param str text: text (if any) to be coloured.
    :param int colour: the colour to be applied to this text and, if no_reset
     is True, all subsequent text until toggled off (if ever).
    :param bool no_reset: colour set kept toggled on so will colour subsequent
     content until toggled off.
    """
    reset_colour = '' if no_reset else cli_conf.DEFAULT_ANSI_COLOUR_BYTE_STR
    for (start, end), inner_colour in cli_conf.BOUNDS2COLOUR.items():
        if start in text:
            if start == cli_conf.LINK_START:
                uon, uoff = "\033[4m", "\033[24m"
            else:
                uon, uoff = '', ''
            text = text.replace(
                start, colourise('', inner_colour, no_reset=True) + uon)
            text = text.replace(
                end, uoff + colourise('', colour, no_reset=True))
    ansi_reverse = cli_conf.REVERSED if reverse else ''
    ansi_bold = cli_conf.BOLD if bold else ''
    text = f"\033[38;5;{colour}m{ansi_bold}{ansi_reverse}{text}{reset_colour}"
    return text

def colourise_low_vis(text):
    return colourise(text, cli_conf.LOW_VIS_COLOUR)

def colourise_plain(text, **_kwargs):
    """
    Useful when a tag is not found.

    Used in context where multiple args supplied but we ignore them here.
    """
    return colourise(text, cli_conf.T)



from ..advisors import filt_block_advisor
from ..ast_funcs import get_assign_name
from .. import code_execution, conf, utils
from ..utils import layout_comment as layout

ASSIGN_TUPLE_XPATH = 'descendant-or-self::Assign/value/Tuple'

@filt_block_advisor(xpath=ASSIGN_TUPLE_XPATH)
def tuple_overview(block_dets, *, repeated_message=False):
    """
    Explain usage of tuples.
    """
    tup_els = block_dets.element.xpath(ASSIGN_TUPLE_XPATH)
    title = layout("""\

        #### Tuple Overview

        """)
    brief_msg = title
    main_msg = title
    name_tups = []
    for tup_el in tup_els:
        name = get_assign_name(tup_el)
        if name is None:  ## no names - e.g. just assignment to other tuples
            continue
        tup = code_execution.get_val(
            block_dets.pre_block_code_str, block_dets.block_code_str, name)
        name_tups.append((name, tup))
        tup_desc = layout(f"""\

            `{name}` is a tuple with {utils.int2nice(len(tup))} items.

            """)
        brief_msg += tup_desc
        main_msg += tup_desc
    if repeated_message:
        extra_msg = ''
    else:
        brief_msg += layout("""\

            Tuples are like lists but the items inside cannot be replaced,
            removed, or added to. For example, if we have a list [1, 2] we can
            append a 3 to it. But if we have a tuple (1, 2) we cannot.

            """)
        main_msg += layout("""\

            Tuples are like lists but they are immutable. That means
            unchangeable.

            """)
        why_immutability_comment = layout("""
            But why would we want a data structure with all those limitations -
            wouldn't a list always be better? In practice it is often useful to
            know that a data structure is not being mutated somewhere inside the
            program. This guarantee makes it easier to reason about what the
            program is doing and what it cannot be doing. We can also use named
            tuples to improve readability.
            """)
        brief_msg += why_immutability_comment
        main_msg += why_immutability_comment
        brief_msg += layout("""\

            Tuples have an order, and can contain duplicate items and items of
            different types (usually not advisable).

            """)

        non_empty_name_tups = [(name, tup) for name, tup in name_tups if tup] 
        if non_empty_name_tups:
            first_name, first_tup = non_empty_name_tups[0]
            tup_replaced = list(first_tup)
            tup_replaced[0] = 100
            tup_replaced = tuple(tup_replaced)
            tup_popped = list(first_tup)
            tup_popped.pop()
            tup_popped = tuple(tup_popped)
            tup_appended = list(first_tup)
            tup_appended.append(3)
            tup_appended = tuple(tup_appended)
            immutability_comment = layout(f"""
                * cannot be *replaced* -
                so we can't run `{first_name}`[0] = 100 to get {tup_replaced}.
                It will raise an exception -
                TypeError: 'tuple' object does not support item assignment)
                * cannot be *removed* -
                so we can't run `{first_name}`.pop() to get {tup_popped}.
                * cannot be *added* - so we can't run `{name}`.append(3) to
                get {tup_appended}.
                """)
        else:
            immutability_comment = layout(f"""
                * cannot be *replaced*. It will raise an exception - TypeError:
                'tuple' object does not support item assignment)
                * cannot be *removed*
                * cannot be *added*
                """)
        main_msg += immutability_comment
        main_msg += layout("""\

            Tuples have an order, and can contain duplicate items and items of
            different types (usually not advisable).

            """)

        friends = ['Selma', 'Willy', 'Principal Skinner']
        family = ['Bart', 'Lisa', 'Marge', 'Homer']
        original_guests = (friends, family)
        guests = (friends + ['Lenny'], family)
        extra_msg = (
            layout(f"""\

                #### GOTCHA - immutable means 100% unchangeable right?

                If you have a tuple of mutable data e.g. lists, the content
                of those lists can be altered even while it remains true that you
                can't replace, remove, or add items inside the tuple.

                Demonstration:

                """)
            +
            layout(f"""\
                friends = {friends}
                family = {family}
                guests = (friends, family)

                """, is_code=True)
            +
            layout(f"""\

                `guests` is now {original_guests}. And because `guests` is
                immutable we can't add, remove, or replace its items. But look
                what happens to the `guests` tuple when we append a person to
                friends:

                """)
            +
            layout(f"""\
                friends.append('Lenny')

                """, is_code=True)
            +
            layout(f"""\

                `guests` is now {guests}. An immutable data structure has
                changed! In reality it is the references to individual items
                that are immutable e.g. a reference to the memory address where
                a mutable list sits.

                #### Named tuples

                Tuples are a light-weight way of defining and storing data. It
                is easier, for example, to write `coord` = (-37, 174) than it is
                to write `coord` = {{'lat': -37, 'lon': 174}}. But it can be
                risky consuming data from tuples in later code if the order or
                number of items changes. What if the tuples you are processing
                change from (lat, lon) to (lon, lat)? References to coord[0]
                would have changed meaning without it being obvious. There can
                be a high risk of bugs.

                Named tuples retain the light-weight advantages of plain vanilla
                tuples but enable us to use a dot notation to reference internal
                items e.g. house_lat = coord.lat. Named tuples also contain
                information on their contents when printed / logged which can
                make debugging much easier.

                Example syntax:

                """)
            +
            layout(f"""\

                from collections import namedtuple
                Coord = namedtuple('Coordinate', 'x, y')
                coord = Coord(-37, 174)
                print(coord)
                # >>> Coordinate(x=-37, y=174))
                """, is_code=True)
            +
            layout(f"""\

                #### "Tupple" vs "Toople"

                There is no consensus on how to pronounce tuple. Should it rhyme
                with cup or with hoop? People from a more mathematical
                background often prefer "toople". Others prefer to follow the
                spelling and say "tupple".
                """)
        )
    message = {
        conf.BRIEF: brief_msg,
        conf.MAIN: main_msg,
        conf.EXTRA: extra_msg,
    }
    return message
